import argparse
from typing import List, Iterable

from seppl import Plugin
from wai.logging import LOGGING_WARNING

from idc.api import DATATYPES, data_type_to_class, ImageData
from idc.api import Reader

READ_ORDER_SEQUENTIAL = "sequential"
READ_ORDER_INTERLEAVED = "interleaved"
READ_ORDERS = [
    READ_ORDER_SEQUENTIAL,
    READ_ORDER_INTERLEAVED,
]


class MultiReader(Reader):

    def __init__(self, readers: List[str] = None, read_order: str = None, data_type: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param readers: the list of base readers to use (command-lines)
        :param read_order: how the base readers are being used
        :type read_order: str
        :param data_type: the type of output to generate from the images
        :type data_type: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.readers = readers
        self.read_order = read_order
        self.data_type = data_type
        self._readers = None
        self._finalize = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "from-multi"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Reads data using the specified base readers and combines their output."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-r", "--reader", type=str, default=None, help="The command-line defining the base reader.", required=True, nargs="+")
        parser.add_argument("-o", "--read_order", choices=READ_ORDERS, type=str, default=READ_ORDER_SEQUENTIAL, help="How to use the output from the readers.", required=False)
        parser.add_argument("-t", "--data_type", choices=DATATYPES, type=str, default=None, help="The type of data to forward", required=True)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.readers = ns.reader
        self.read_order = ns.read_order
        self.data_type = ns.data_type

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        if self.data_type is None:
            return [ImageData]
        else:
            return [data_type_to_class(self.data_type)]

    def _parse_commandline(self, cmdline: str) -> List[Plugin]:
        """
        Parses the command-line and returns the list of plugins it represents.
        Raises an exception in case of an invalid sub-flow.

        :param cmdline: the command-line to parse
        :type cmdline: str
        :return:
        """
        from idc.registry import available_readers
        from seppl import args_to_objects, split_args, split_cmdline

        # split command-line into valid plugin subsets
        valid = available_readers()
        args = split_args(split_cmdline(cmdline), list(valid.keys()))
        return args_to_objects(args, valid, allow_global_options=False)

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.data_type is None:
            raise Exception("No data type defined!")
        if (self.readers is None) or (len(self.readers) == 0):
            raise Exception("No reader(s) defined!")
        if self.read_order is None:
            self.read_order = READ_ORDER_SEQUENTIAL
        if self.read_order not in READ_ORDERS:
            raise Exception("Unknown read order: %s" % self.read_order)
        self._readers = []
        for reader in self.readers:
            objs = self._parse_commandline(reader)
            if len(objs) == 1:
                _reader = objs[0]
                if _reader.generates() == self.generates():
                    self._readers.append(_reader)
                else:
                    raise Exception("Reader '%s' generates '%s' but '%s' is required!" % (reader, str(_reader.generates()), str(self.generates())))
            else:
                raise Exception("Failed to obtain a single reader from command-line: %s" % reader)
        for reader in self._readers:
            reader.initialize()
            reader.session = self.session
        self.logger().info("# readers: %d" % len(self._readers))
        self._finalize = []

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        if self.read_order == READ_ORDER_SEQUENTIAL:
            readers = self._readers[:]
            for reader in readers:
                while not reader.has_finished():
                    for data in reader.read():
                        yield data
                self._finalize.append(self._readers.pop(0))
        elif self.read_order == READ_ORDER_INTERLEAVED:
            while len(self._readers) > 0:
                for reader in self._readers:
                    if not reader.has_finished():
                        for data in reader.read():
                            yield data
                    if reader.has_finished():
                        self._readers.remove(reader)
                        self._finalize.append(reader)
        else:
            raise Exception("Unhandled read order: %s" % self.read_order)

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return len(self._readers) == 0

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        super().finalize()
        if self._finalize is not None:
            for reader in self._finalize:
                reader.finalize()



import argparse
import os
from datetime import datetime
from typing import List

from wai.logging import LOGGING_WARNING
from idc.api import ObjectDetectionData, SplittableStreamWriter, get_object_label, make_list, AnnotationsOnlyWriter, add_annotations_only_param
from opex import ObjectPredictions, ObjectPrediction, BBox, Polygon
from seppl.placeholders import placeholder_list, InputBasedPlaceholderSupporter


class OPEXObjectDetectionWriter(SplittableStreamWriter, AnnotationsOnlyWriter, InputBasedPlaceholderSupporter):

    def __init__(self, output_dir: str = None, annotations_only: bool = None,
                 split_names: List[str] = None, split_ratios: List[int] = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the writer.

        :param output_dir: the output directory to save the image/report in
        :type output_dir: str
        :param annotations_only: whether to output only the annotations and not the images
        :type annotations_only: bool
        :param split_names: the names of the splits, no splitting if None
        :type split_names: list
        :param split_ratios: the integer ratios of the splits (must sum up to 100)
        :type split_ratios: list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(split_names=split_names, split_ratios=split_ratios, logger_name=logger_name, logging_level=logging_level)
        self.output_dir = output_dir
        self.annotations_only = annotations_only

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "to-opex-od"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Saves the bounding box/polygon definitions in OPEX .json format."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-o", "--output", type=str, help="The directory to store the images/.json files in. Any defined splits get added beneath there. " + placeholder_list(obj=self), required=True)
        add_annotations_only_param(parser)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.output_dir = ns.output
        self.annotations_only = ns.annotations_only

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.annotations_only is None:
            self.annotations_only = False

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        for item in make_list(data):
            sub_dir = self.session.expand_placeholders(self.output_dir)
            if self.splitter is not None:
                split = self.splitter.next(item=item.image_name)
                sub_dir = os.path.join(sub_dir, split)
            if not os.path.exists(sub_dir):
                self.logger().info("Creating dir: %s" % sub_dir)
                os.makedirs(sub_dir)

            absolute = None
            if item.has_annotation():
                absolute = item.get_absolute()

            # image
            path = sub_dir
            os.makedirs(path, exist_ok=True)
            path = os.path.join(path, item.image_name)
            if not self.annotations_only:
                self.logger().info("Writing image to: %s" % path)
                item.save_image(path)

            # annotations
            if absolute is not None:
                objs = []
                for obj in absolute:
                    # score
                    score = 0.0
                    if "score" in obj.metadata:
                        try:
                            score = float(obj.metadata["score"])
                        except:
                            pass
                    # label
                    label = get_object_label(obj)
                    # meta data
                    meta = dict()
                    for k, v in obj.metadata.items():
                        if k in ["score", "type"]:
                            continue
                        meta[k] = str(v)
                    if len(meta) == 0:
                        meta = None
                    # bbox
                    bbox = BBox(left=obj.x, top=obj.y, right=obj.x + obj.width - 1, bottom=obj.y + obj.height - 1)
                    # polygon
                    points = []
                    if obj.has_polygon():
                        for x, y in zip(obj.get_polygon_x(), obj.get_polygon_y()):
                            points.append([x, y])
                    else:
                        points.append([obj.x, obj.y])
                        points.append([obj.x + obj.width - 1, obj.y])
                        points.append([obj.x + obj.width - 1, obj.y + obj.height - 1])
                        points.append([obj.x, obj.y + obj.height - 1])
                    polygon = Polygon(points=points)
                    objs.append(ObjectPrediction(score=score, label=label, bbox=bbox, polygon=polygon, meta=meta))

                # meta
                meta = dict()
                if item.has_metadata():
                    for k, v in item.get_metadata().items():
                        if k.startswith("Parent ID["):
                            continue
                        meta[k] = str(v)
                if len(meta) == 0:
                    meta = None

                # timestamp
                ts = datetime.now().strftime("%Y%m%d_%H%M%S.%f")

                # ID
                if item.image_name is None:
                    id = ts
                else:
                    id = os.path.splitext(os.path.basename(item.image_name))[0]

                # assemble predictions
                if meta is None:
                    preds = ObjectPredictions(id=id, timestamp=ts, objects=objs)
                else:
                    preds = ObjectPredictions(id=id, timestamp=ts, objects=objs, meta=meta)

                path = sub_dir
                os.makedirs(path, exist_ok=True)
                path = os.path.join(path, item.image_name)
                path = os.path.splitext(path)[0] + ".json"
                self.logger().info("Writing annotations to: %s" % path)
                preds.save_json_to_file(path)



import argparse
from typing import List

from seppl import AnyData
from seppl.io import Filter
from seppl.placeholders import add_placeholder, InputBasedPlaceholderSupporter, placeholder_list
from wai.logging import LOGGING_WARNING

from idc.api import ImageData


class SetPlaceholder(Filter, InputBasedPlaceholderSupporter):

    def __init__(self, placeholder: str = None, value: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param placeholder: the name of the placeholder (without curly brackets)
        :type placeholder: str
        :param value: the value of the placeholder
        :type value: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.placeholder = placeholder
        self.value = value

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "set-placeholder"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Sets the placeholder to the specified value when data passes through. The value can contain other placeholders, which get expanded each time data passes through."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-p", "--placeholder", type=str, help="The name of the placeholder, without curly brackets.", default=None, required=True)
        parser.add_argument("-v", "--value", type=str, help="The value of the placeholder, may contain other placeholders. " + placeholder_list(obj=self), default=None, required=True)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.placeholder = ns.placeholder
        self.value = ns.value

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.placeholder is None:
            raise Exception("No placeholder name provided!")
        if self.value is None:
            raise Exception("No placeholder value provided!")

    def _do_process(self, data: ImageData):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        value = self.session.expand_placeholders(self.value)
        self.logger().info("%s -> %s" % (self.placeholder, value))
        add_placeholder(self.placeholder, "no description", False, lambda i: value)
        return data



import argparse
from typing import List, Iterable, Union

from seppl.placeholders import PlaceholderSupporter, placeholder_list
from seppl.io import locate_files
from wai.logging import LOGGING_WARNING

from idc.api import ImageSegmentationData, locate_file, load_image_from_file, from_grayscale, JPEG_EXTENSIONS
from idc.api import Reader


class GrayscaleImageSegmentationReader(Reader, PlaceholderSupporter):

    def __init__(self, source: Union[str, List[str]] = None, source_list: Union[str, List[str]] = None,
                 image_path_rel: str = None, labels: List[str] = None, background: int = None, resume_from: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param source: the filename(s)
        :param source_list: the file(s) with filename(s)
        :param image_path_rel: the relative path from the annotations to the images
        :type image_path_rel: str
        :param labels: the list of labels
        :type labels: list
        :param background: the index (0-255) that is used as background
        :type background: int
        :param resume_from: the file to resume from (glob)
        :type resume_from: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.source = source
        self.source_list = source_list
        self.image_path_rel = image_path_rel
        self.labels = labels
        self.background = background
        self.resume_from = resume_from
        self._label_mapping = None
        self._inputs = None
        self._current_input = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "from-grayscale-is"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Loads the annotations from associated grayscale PNG files."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--input", type=str, help="Path to the PNG file(s) to read; glob syntax is supported; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("-I", "--input_list", type=str, help="Path to the text file(s) listing the text PNG to use; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("--resume_from", type=str, help="Glob expression matching the file to resume from, e.g., '*/012345.png'", required=False)
        parser.add_argument("--image_path_rel", metavar="PATH", type=str, default=None, help="The relative path from the annotations to the images directory", required=False)
        parser.add_argument("--labels", metavar="LABEL", type=str, default=None, help="The labels that the indices represent.", nargs="+")
        parser.add_argument("--background", type=int, help="The index (0-255) that is used for the background", required=False, default=0)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.source = ns.input
        self.source_list = ns.input_list
        self.image_path_rel = ns.image_path_rel
        self.labels = ns.labels
        self.background = ns.background
        self.resume_from = ns.resume_from

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ImageSegmentationData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.labels is None:
            raise Exception("No labels defined!")
        self._inputs = locate_files(self.source, input_lists=self.source_list, fail_if_empty=True, default_glob="*.png", resume_from=self.resume_from)
        self._label_mapping = dict()
        for i, label in enumerate(self.labels):
            self._label_mapping[i] = label
        self.logger().debug("label mapping: %s" % str(self._label_mapping))

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        self._current_input = self._inputs.pop(0)
        self.session.current_input = self._current_input

        # associated images?
        imgs = locate_file(self.session.current_input, JPEG_EXTENSIONS, rel_path=self.image_path_rel)
        if len(imgs) == 0:
            self.logger().warning("Failed to locate associated image for: %s" % self.session.current_input)
            yield None

        # read annotations
        self.logger().info("Reading from: " + str(self.session.current_input))
        ann = load_image_from_file(self.session.current_input)
        annotations = from_grayscale(ann, self.labels, self._label_mapping, self.logger(), background=self.background)

        # associated image
        if len(imgs) > 1:
            self.logger().warning("Found more than one image associated with annotation, using first: %s" % imgs[0])
            yield None

        yield ImageSegmentationData(source=imgs[0], annotation=annotations)

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return len(self._inputs) == 0



import argparse
import copy
import numpy as np
import re
from shapely.geometry import Polygon
from typing import List, Dict

from wai.logging import LOGGING_WARNING
from wai.common.adams.imaging.locateobjects import LocatedObjects, LocatedObject, normalized_to_absolute, \
    NormalizedLocatedObjects
from seppl.io import Filter
from idc.api import ObjectDetectionData, ImageClassificationData, ImageSegmentationData, locatedobject_polygon_to_shapely, \
    intersect_over_union, get_object_label, flatten_list, make_list


class FilterLabels(Filter):
    """
    Filters out labels according to the parameters.
    """

    def __init__(self, labels: List[str] = None, regexp: str = None,
                 region: str = None, min_iou: float = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param labels: the labels to use
        :type labels: list
        :param regexp: the regular expression for using only a subset
        :type regexp: str
        :param region: region that the object must overlap with in order to be included (object detection only). Between 0-1 the values are considered normalized, otherwise absolute pixels. format: x,y,w,h
        :type region: str
        :param min_iou: the minimum IoU (intersect over union) that the object must have with the region in order to be considered an overlap (object detection only)
        :type min_iou: float
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.labels = labels
        self.regexp = regexp
        self.region = region
        self.min_iou = min_iou
        self._pattern = None
        self._region = None
        self._normalized = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "filter-labels"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Filters out labels according to the parameters."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData, ImageClassificationData, ImageSegmentationData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData, ImageClassificationData, ImageSegmentationData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("--labels", type=str, default=None, help="The labels to use", required=False, nargs="*")
        parser.add_argument("--regexp", type=str, default=None, help="Regular expression for using only a subset of labels", required=False)
        parser.add_argument("--region", type=str, default=None, metavar="x,y,w,h", help="Region that the object must overlap with in order to be included (object detection only). Between 0-1 the values are considered normalized, otherwise absolute pixels.", required=False)
        parser.add_argument("--min_iou", type=float, default=None, help="The minimum IoU (intersect over union) that the object must have with the region in order to be considered an overlap (object detection only)", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.labels = ns.labels
        self.regexp = ns.regexp
        self.region = ns.region
        self.min_iou = ns.min_iou

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()

        if self.labels is None:
            self.labels = []

        self._pattern = re.compile(self.regexp) if (self.regexp is not None) else None

        if self.region is None:
            self._region = None
            self._normalized = False
        else:
            parts = self.region.split(",")
            if len(parts) != 4:
                raise Exception("Region must have format 'x,y,w,h', but found: %s" % self.region)
            self._region = [float(x) for x in parts]
            self._normalized = sum([(0 if (x < 1) else 1) for x in self._region]) < 4

    def remove_invalid_objects(self, located_objects: LocatedObjects, width: int, height: int) -> LocatedObjects:
        """
        Removes objects with labels that are not valid under the given options.

        :param located_objects: The located objects to process.
        :param width: the width of the image
        :type width: int
        :param height: the height of the image
        :type height: int
        :return: the (potentially) updated objects
        :rtype: LocatedObjects
        """
        # Create a list of objects to remove
        invalid_objects: List[int] = []

        if isinstance(located_objects, NormalizedLocatedObjects):
            absolute = normalized_to_absolute(located_objects, width, height)
        else:
            absolute = located_objects

        # Search the located objects
        for index, located_object in enumerate(absolute):
            if not self.filter_object(located_object):
                invalid_objects.append(index)
            elif not self.filter_region(located_object, width, height):
                invalid_objects.append(index)

        # Remove the invalid objects
        if len(invalid_objects) > 0:
            located_objects = copy.deepcopy(located_objects)
            for index in reversed(invalid_objects):
                located_objects.pop(index)

        return located_objects

    def filter_label(self, label: str) -> bool:
        """
        Filter function which selects labels that match the given options.

        :param label:   The label to test.
        :return:        True if the label matches, false if not.
        """
        if len(self.labels) == 0 and self._pattern is None:
            return True
        elif len(self.labels) == 0:
            return bool(self._pattern.match(label))
        elif self._pattern is None:
            return label in self.labels
        else:
            return bool(self._pattern.match(label)) or label in self.labels

    def filter_object(self, located_object: LocatedObject) -> bool:
        """
        Filter function which selects objects whose labels match the given options.

        :param located_object:  The located object to test.
        :return:                True if the object matches, false if not.
        """
        # Filter the label
        label = get_object_label(located_object)
        return self.filter_label(label)

    def filter_region(self, located_object: LocatedObject, width: int, height: int) -> bool:
        """
        Ensures that the object fits into the defined region.

        :param located_object: the object to check
        :param width: the width of the image
        :type width: int
        :param height: the height of the image
        :type height: int
        :return: True if the objects overlaps the region or no region defined at all
        """
        if self._region is None:
            return True

        if self._normalized:
            x = int(self._region[0] * width)
            y = int(self._region[1] * height)
            w = int(self._region[2] * width)
            h = int(self._region[3] * height)
        else:
            x = self._region[0]
            y = self._region[1]
            w = self._region[2]
            h = self._region[3]

        object_poly = locatedobject_polygon_to_shapely(located_object)
        region_poly = Polygon([(x, y), (x+w-1, y), (x+w-1, y+h-1), (x, y+h-1)])
        iou = intersect_over_union(object_poly, region_poly)

        return iou > self.min_iou

    def remove_invalid_layers(self, layers: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        """
        Removes all invalid layers.

        :param layers: the layers to process
        :type layers: dict
        :return: the updated layers
        :rtype: dict
        """
        delete = []
        for label in layers:
            if not self.filter_label(label):
                delete.append(label)
        if len(delete) > 0:
            layers = copy.deepcopy(layers)
            for label in delete:
                del layers[label]
        return layers

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            if isinstance(item, ObjectDetectionData):
                # Use the options to filter the located objects by label
                objects_new = self.remove_invalid_objects(item.annotation, item.image_width, item.image_height)
                # no annotations left? mark as negative
                if len(objects_new) == 0:
                    item_new = item.duplicate()
                    item_new.annotation = None
                    result.append(item_new)
                else:
                    item_new = item.duplicate()
                    item_new.annotation = objects_new
                    result.append(item_new)
            elif isinstance(item, ImageClassificationData):
                # mark as negative if label doesn't match
                if not self.filter_label(item.annotation):
                    item_new = item.duplicate()
                    item_new.annotation = None
                    result.append(item_new)
                else:
                    result.append(item)
            elif isinstance(item, ImageSegmentationData):
                # remove layers
                layers = self.remove_invalid_layers(item.annotation.layers)
                # no layers left? mark as negative
                if len(layers) == 0:
                    item_new = item.duplicate()
                    item_new.annotation = None
                    result.append(item_new)
                else:
                    item_new = item.duplicate()
                    item_new.annotation.layers = layers
                    result.append(item_new)
            else:
                self.logger().warning("Cannot process data type: %s" % str(type(item)))

        return flatten_list(result)



from ._coerce_box import CoerceBox
from ._coerce_mask import CoerceMask
from ._dimension_discarder import DimensionDiscarder
from ._label_present import LabelPresent
from ._polygon_discarder import PolygonDiscarder
from ._polygon_simplifier import PolygonSimplifier



import argparse
import copy
from typing import List, Tuple

from shapely import simplify, Polygon
from wai.logging import LOGGING_WARNING
from wai.common.adams.imaging.locateobjects import LocatedObjects, LocatedObject
from wai.common.geometry import Point as WaiPoint, Polygon as WaiPolygon

from seppl.io import Filter
from idc.api import ObjectDetectionData, flatten_list, make_list, locatedobject_polygon_to_shapely


class PolygonSimplifier(Filter):
    """
    Simplifies polygons according to the tolerance parameter: the smaller the tolerance, the closer to the original.
    """

    def __init__(self, tolerance: float = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param tolerance: the tolerance parameter to use, the smaller, the closer to the original
        :type tolerance: float
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.tolerance = tolerance

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "polygon-simplifier"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Simplifies polygons according to the tolerance parameter: the smaller the tolerance, the closer to the original."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-t", "--tolerance", type=float, default=0.01, help="The tolerance for the simplification.", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.tolerance = ns.tolerance

    def _simplify(self, obj: LocatedObject) -> Tuple[LocatedObject, bool]:
        """
        Simplifies the polygon of the object.

        :param obj: the object to process
        :type obj: LocatedObject
        :return: the potentially updated object
        :rtype: tuple
        """
        result = False

        if obj.has_polygon():
            spoly = locatedobject_polygon_to_shapely(obj)
            spoly_new = simplify(spoly, self.tolerance)
            if isinstance(spoly_new, Polygon) and (len(spoly_new.exterior.coords) < len(spoly.exterior.coords)):
                result = True
                obj = copy.deepcopy(obj)
                x_list, y_list = spoly_new.exterior.coords.xy
                points = []
                for i in range(len(x_list)):
                    points.append(WaiPoint(x=x_list[i], y=y_list[i]))
                obj.set_polygon(WaiPolygon(*points))

        return obj, result

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            ann = LocatedObjects()
            count = 0
            for obj in item.annotation:
                new_obj, updated = self._simplify(obj)
                ann.append(new_obj)
                if updated:
                    count += 1
            if count > 0:
                item = item.duplicate(annotation=ann)
                self.logger().info("%s: %d polygons updated" % (item.image_name, count))
            result.append(item)

        return flatten_list(result)



import base64
import copy
import io
import logging
import os.path
import shutil
from typing import Dict, Optional, Tuple, List, Union

import imagesize
import numpy as np
from PIL import Image
from image_complete.bmp import is_bmp
from image_complete.jpg import is_jpg
from image_complete.png import is_png
from seppl import MetaDataHandler, LoggingHandler
from ._utils import load_image_from_bytes, safe_deepcopy
from wai.logging import set_logging_level, LOGGING_INFO

_logger = None

FORMAT_JPEG = "JPEG"
FORMAT_PNG = "PNG"
FORMAT_BMP = "BMP"
FORMATS = [
    FORMAT_JPEG,
    FORMAT_PNG,
    FORMAT_BMP,
]
FORMAT_EXTENSIONS = {
    FORMAT_JPEG: ".jpg",
    FORMAT_PNG: ".png",
    FORMAT_BMP: ".bmp",
}

IDC_JPEG_QUALITY = "IDC_JPEG_QUALITY"
""" the environment variable to overriding the default JPEG quality. """

DEFAULT_JPEG_QUALITY = 90
""" the default quality to use when saving jpg/jpeg files. """

JPEG_QUALITY = None
""" the quality to use when writing JPEG files. """


def logger() -> logging.Logger:
    """
    Returns the logger instance to use, initializes it if necessary.

    :return: the logger instance
    :rtype: logging.Logger
    """
    global _logger
    if _logger is None:
        _logger = logging.getLogger("idc.api.data")
        set_logging_level(_logger, LOGGING_INFO)
    return _logger


def jpeg_quality() -> int:
    """
    Returns the quality to use for jpeg images.

    :return: the quality
    :rtype: int
    """
    global JPEG_QUALITY
    if JPEG_QUALITY is None:
        try:
            JPEG_QUALITY = DEFAULT_JPEG_QUALITY
            if IDC_JPEG_QUALITY in os.environ:
                JPEG_QUALITY = int(os.getenv(IDC_JPEG_QUALITY, str(DEFAULT_JPEG_QUALITY)))
                logger().info("Using JPEG quality: %d%%" % JPEG_QUALITY)
        except:
            JPEG_QUALITY = DEFAULT_JPEG_QUALITY
    return JPEG_QUALITY


def save_image(img: Image.Image, path: str, make_dirs: bool = False):
    """
    Saves the image in the specified location.
    For JPEG images, takes the quality into account.

    :param img: the image to save
    :type img: Image.Image
    :param path: the path to save the image to
    :type path: str
    :param make_dirs: whether to create parent dirs if necessary
    :type make_dirs: bool
    """
    if make_dirs:
        parent_dir = os.path.dirname(path)
        if not os.path.exists(parent_dir):
            os.makedirs(parent_dir)
    if img.format == "JPEG":
        img.save(path, quality=jpeg_quality())
    else:
        img.save(path)


def array_to_image(array: Union[np.ndarray, Image.Image], image_format: str) -> Tuple[Image.Image, io.BytesIO]:
    """
    Turns the numpy array back into an image of the specified format.
    Returns the image data structure and the bytes representing it.
    For JPEG images, takes the quality into account.

    :param array: the Image/array to convert
    :type array: np.ndarray or Image.Image
    :param image_format: the image format to generate
    :type image_format: str
    :return: the generated image data structure
    :rtype: tuple
    """
    if isinstance(array, Image.Image):
        img = array
    else:
        img = Image.fromarray(np.uint8(array))
    img_bytes = io.BytesIO()
    if image_format == "JPEG":
        img.save(img_bytes, format=image_format, quality=jpeg_quality())
    else:
        img.save(img_bytes, format=image_format)
    return img, img_bytes


def empty_image(mode: str, width: int, height: int, image_format: str) -> Tuple[Image.Image, io.BytesIO]:
    """
    Creates an empty image and returns the image and the bytes representation.

    :param mode: the image mode, e.g., RGB
    :type mode: str
    :param width: the width of the image
    :type width: int
    :param height: the height of the image
    :type height: int
    :param image_format: the image format to use, e.g., JPEG
    :type image_format: str
    :return: the tuple of image and bytes
    :rtype: tuple
    """
    img = Image.new(mode, (width, height))
    img_bytes = io.BytesIO()
    if image_format == "JPEG":
        img.save(img_bytes, format=image_format, quality=jpeg_quality())
    else:
        img.save(img_bytes, format=image_format)
    return img, img_bytes


class ImageData(MetaDataHandler, LoggingHandler):

    def __init__(self, source: str = None, image_name: str = None, data: bytes = None,
                 image: Image.Image = None, image_format: str = None, image_size: Tuple[int, int] = None,
                 metadata: Dict = None, annotation=None):
        self._logger = None
        """ for logging. """
        self._source = source
        """ the full path to the image file. """
        self._image_name = image_name
        """ the name of the image file (no path). """
        self._data = data
        """ the binary image data. """
        self._image = image
        """ the Pillow image. """
        self._image_format = image_format
        """ the format of the image. """
        self._image_size = image_size
        """ the size (width, height) tuple of the image. """
        self._metadata = metadata
        """ the dictionary with optional meta-data. """
        self.annotation = annotation
        """ the associated annotation data. """

    def logger(self) -> logging.Logger:
        """
        Returns the logger instance to use.

        :return: the logger
        :rtype: logging.Logger
        """
        if self._logger is None:
            self._logger = logging.getLogger(self.__class__.__name__)
        return self._logger

    @property
    def source(self) -> Optional[str]:
        """
        Returns the source filename.

        :return: the full filename, if available
        :rtype: str
        """
        return self._source

    @property
    def image(self) -> Image.Image:
        """
        Returns the image, loads it if necessary.

        :return: the pillow image data structure, None if not available or failed to load
        :rtype: Image.Image
        """
        if self._image is not None:
            return self._image
        if self._data is not None:
            self._image = load_image_from_bytes(self._data)
            self._image_format = self._image.format
            return self._image
        if self._source is not None:
            self._image_name = os.path.basename(self._source)
            self._image = Image.open(self._source)
            self._image_format = self._image.format
            return self._image
        return None

    @property
    def image_bytes(self):
        """
        Turns the pillow image into bytes using the current image format.

        :return: the generated bytes
        """
        buffer = io.BytesIO()
        if self.image is not None:
            self.image.save(buffer, format=self.image_format)
        return buffer.getvalue()

    @property
    def image_name(self) -> Optional[str]:
        """
        Returns the name of the image.

        :return: the image name, can be None
        :rtype: str
        """
        if self._image_name is not None:
            return self._image_name
        elif self.source is not None:
            return os.path.basename(self.source)
        else:
            return None

    @image_name.setter
    def image_name(self, s: str):
        """
        Sets the new name.

        :param s: the new name
        :type s: str
        """
        self._image_name = s

    @property
    def image_format(self) -> Optional[str]:
        """
        Returns the format of the image.

        :return: the image format, can be None
        :rtype: str
        """
        if self._image_format is None:
            if self.source is not None:
                to_check = self.source
            else:
                to_check = self.data
            if to_check is not None:
                if is_jpg(to_check):
                    self._image_format = FORMAT_JPEG
                elif is_png(to_check):
                    self._image_format = FORMAT_PNG
                elif is_bmp(to_check):
                    self._image_format = FORMAT_BMP
        if self._image_format is None:
            if self.image is None:
                return None
        return self._image_format

    @property
    def image_size(self) -> Optional[Tuple[int, int]]:
        """
        Returns the size tuple (width, height) for the image.

        :return: the width/height tuple, None if failed to determine
        :rtype: tuple
        """
        if self._image_size is not None:
            return self._image_size

        if self._data is not None:
            try:
                self._image_size = imagesize.get(self._data)
                return self._image_size
            except:
                pass

        if self._source is not None:
            try:
                self._image_size = imagesize.get(self._source)
                return self._image_size
            except:
                pass

        if self.image is not None:
            return self.image.size

        return None

    @property
    def image_width(self) -> Optional[int]:
        """
        Returns the width of the image.

        :return: the width, None if failed to determine
        :rtype: int
        """
        size = self.image_size
        if size is None:
            return None
        else:
            return size[0]

    @property
    def image_height(self) -> Optional[int]:
        """
        Returns the height of the image.

        :return: the height, None if failed to determine
        :rtype: int
        """
        size = self.image_size
        if size is None:
            return None
        else:
            return size[1]

    @property
    def data(self) -> bytes:
        """
        Returns the internal data, if any.

        :return: the data
        """
        return self._data

    @data.setter
    def data(self, data: bytes):
        """
        Uses the provided data.

        :param data: the data to use
        """
        self._image_name = self.image_name
        self._source = None
        self._image_format = None
        self._image_size = None
        self._data = data

    def save_image(self, path: str, make_dirs: bool = False) -> bool:
        """
        Saves the image under the specified path.

        :param path: the path to save the image under
        :type path: str
        :param make_dirs: whether to create any missing parent dirs
        :type make_dirs: bool
        :return: whether the file was saved
        :rtype: bool
        """
        if self._source is not None:
            if os.path.exists(path) and os.path.exists(self._source) and os.path.samefile(path, self._source):
                self.logger().warning("Input/output image are the same, skipping!")
                return False
        if make_dirs:
            parent_dir = os.path.dirname(path)
            if not os.path.exists(parent_dir):
                self.logger().info("Creating dir: %s" % parent_dir)
                os.makedirs(parent_dir)
        if (self._data is None) and (self._source is not None) and (os.path.exists(self._source)):
            shutil.copy(self._source, path)
            return True
        if self._image is not None:
            save_image(self._image, path)
            return True
        if self._data is not None:
            with open(path, "wb") as fp:
                fp.write(self._data)
            return True
        return False

    def has_annotation(self) -> bool:
        """
        Checks whether annotations are present.

        :return: True if annotations present
        :rtype: bool
        """
        return self.annotation is not None

    def has_metadata(self) -> bool:
        """
        Returns whether meta-data is present.

        :return: True if meta-data present
        :rtype: bool
        """
        return self._metadata is not None

    def get_metadata(self) -> Optional[Dict]:
        """
        Returns the meta-data.

        :return: the meta-data, None if not available
        :rtype: dict
        """
        return self._metadata

    def set_metadata(self, metadata: Optional[Dict]):
        """
        Sets the meta-data to use.

        :param metadata: the new meta-data, can be None
        :type metadata: dict
        """
        self._metadata = metadata

    def duplicate(self, source: str = None, force_no_source: bool = None,
                  name: str = None, data: bytes = None,
                  image: Image.Image = None, image_format: str = None,
                  size: Tuple[int, int] = None,
                  metadata: Dict = None, annotation=None):
        """
        Duplicates the container overwriting existing data with any provided data.

        :param source: the source to use
        :type source: str
        :param force_no_source: if True, then source is set to None
        :type force_no_source: bool
        :param name: the name to use
        :type name: str
        :param data: the data to use
        :type data: bytes
        :param image: the Pillow image to use
        :type image: Image.Image
        :param image_format: the image format
        :type image_format: str
        :param size: the size tuple
        :type size: tuple
        :param metadata: the metadata
        :type metadata: dict
        :param annotation: the annotations
        :return: the duplicated container
        """
        if (force_no_source is not None) and force_no_source:
            source = None
        else:
            if source is None:
                source = self._source
        if name is None:
            name = self._image_name
        if data is None:
            data = safe_deepcopy(self._data)
        # if the source changes, we need to force loading the image
        if ((image is None) and (self._image is not None)) or (source != self._source):
            image = copy.deepcopy(self.image)
        if image_format is None:
            image_format = self._image_format
        if size is None:
            size = safe_deepcopy(self._image_size)
        if metadata is None:
            metadata = safe_deepcopy(self._metadata)
        if annotation is None:
            annotation = safe_deepcopy(self.annotation)

        return type(self)(source=source, image_name=name, data=data,
                          image=image, image_format=image_format, image_size=size,
                          metadata=metadata, annotation=annotation)

    def _annotation_to_dict(self):
        """
        Turns the annotations into a dictionary.

        :return: the generated dictionary
        :rtype: dict
        """
        raise NotImplementedError()

    def to_dict(self, source: bool = True, image: bool = True, annotation: bool = True, metadata: bool = True):
        """
        Returns itself as a dictionary that can be saved as JSON.

        :param source: whether to include the source
        :type source: bool
        :param image: whether to include the image
        :type image: bool
        :param annotation: whether to include the annotations
        :type annotation: bool
        :param metadata: whether to include the metadata
        :type metadata: bool
        :return: the generated dictionary
        :rtype: dict
        """
        result = dict()
        if source and (self.source is not None):
            result["source"] = self.source
        if self.image_name is not None:
            result["name"] = self.image_name
        if self.image_format is not None:
            result["format"] = self.image_format
        if self.image_width is not None:
            result["width"] = self.image_width
        if self.image_height is not None:
            result["height"] = self.image_height
        if image:
            result["image"] = base64.encodebytes(self.image_bytes).decode("ascii")
        if annotation and (self.annotation is not None):
            result["annotation"] = self._annotation_to_dict()
        if metadata and (self.get_metadata() is not None):
            result["metadata"] = copy.deepcopy(self.get_metadata())
        return result


def make_list(data, cls=ImageData) -> List:
    """
    Wraps the data item in a list if not already a list.

    :param data: the data item to wrap if necessary
    :param cls: the type of class to check for
    :return: the list
    :rtype: list
    """
    if isinstance(data, cls):
        data = [data]
    return data


def flatten_list(data: List):
    """
    If the list contains only a single item, then it returns that instead of a list.

    :param data: the list to check
    :type data: list
    :return: the list or single item
    """
    if len(data) == 1:
        return data[0]
    else:
        return data



import copy
import csv
import importlib
import inspect
import io
import logging
import numpy as np
import os
from typing import Optional, Union, List, Dict, Tuple, Callable, Any

from PIL import Image


JPEG_EXTENSIONS = [".jpg", ".jpeg", ".JPG", ".JPEG"]


def strip_suffix(path: str, suffix: str) -> str:
    """
    Removes the suffix from the file, if possible.

    :param path: the filename to process
    :type path: str
    :param suffix: the suffix to remove (including extension); ignored if None or ""
    :type suffix: str
    :return: the (potentially) updated filename
    :rtype: str
    """
    if suffix is not None:
        if len(suffix) == 0:
            suffix = None
    if suffix is not None:
        if path.endswith(suffix):
            return path[0:-len(suffix)]
    return path


def locate_file(path: str, ext: Union[str, List[str]], rel_path: str = None, suffix: str = None) -> List[str]:
    """
    Tries to locate the associate files for the given path by replacing its extension by the provided ones.

    :param path: the base path to use
    :type path: str
    :param ext: the extension(s) to look for (incl dot)
    :type ext: str or list
    :param suffix: the suffix to strip from the files, ignored if None or ""
    :type suffix: str
    :param rel_path: the relative path to the annotation to use for looking for associated files, ignored if None
    :type rel_path: str
    :return: the located files
    :rtype: list
    """
    result = []
    if rel_path is not None:
        parent_path = os.path.dirname(path)
        name = os.path.basename(path)
        path = os.path.join(parent_path, rel_path, name)
    path = strip_suffix(path, suffix)
    noext = os.path.splitext(path)[0]
    for current in ext:
        path = noext + current
        if os.path.exists(path):
            result.append(path)
    return result


def locate_image(path: str, rel_path: str = None, suffix: str = None) -> Optional[str]:
    """
    Tries to locate the image (png or jpg) for the given path by replacing its extension.

    :param path: the base path to use
    :type path: str
    :param rel_path: the relative path to the annotation to use for looking for images, ignored if None
    :type rel_path: str
    :param suffix: the suffix to strip from the files, ignored if None or ""
    :type suffix: str
    :return: the located image, None if not found
    :rtype: str
    """
    ext = [".png", ".PNG", ".jpg", ".JPG", ".jpeg", ".JPEG"]
    images = locate_file(path, ext, rel_path=rel_path, suffix=suffix)
    if len(images) == 0:
        return None
    else:
        return images[0]


def load_image_from_bytes(data) -> Image:
    """
    Loads a Pillow image from the bytes.

    :param data: the bytes to load from
    :return: the image loaded from the data
    :rtype: Image
    """
    return Image.open(io.BytesIO(data))


def load_image_from_file(path: str) -> Image:
    """
    Loads a Pillow image from the specified file.

    :param path: the path to load from
    :return: the image loaded from the file
    :rtype: Image
    """
    return Image.open(path)


def load_labels(path: str, logger: logging.Logger = None) -> Tuple[List[str], Dict[int, str]]:
    """
    Loads the comma-separated labels from the text file and returns
    them as list and as index/label mapping.

    :param path: the file to load the labels from
    :type path: str
    :param logger: the optional logger to use for outputting information
    :type logger: logging.Logger
    :return: the tuple of labels list and dictionary of index/label mapping
    :rtype: tuple
    """
    if logger is not None:
        logger.info("Reading labels from: %s" % str(path))
    with open(path, "r") as fp:
        line = fp.readline()
    labels = [x.strip() for x in line.strip().split(",")]
    label_mapping = dict()
    for i, label in enumerate(labels):
        label_mapping[i] = label
    if logger is not None:
        logger.debug("label mapping: %s" % str(label_mapping))
    return labels, label_mapping


def save_labels(path: str, labels: List[str], logger: logging.Logger = None):
    """
    Writes the labels as comma-separated list to the specified file.

    :param path: the file to write the labels to
    :type path: str
    :param labels: the labels to write
    :type labels: list
    :param logger: the optional logger to use for outputting information
    :type logger: logging.Logger
    """
    if logger is not None:
        logger.info("Writing labels file: %s" % path)
    with open(path, "w") as fp:
        fp.write(",".join(labels))


def save_labels_csv(path: str, labels: Dict[int, str], logger: logging.Logger = None):
    """
    Writes the labels as CSV (Index,Label) to the specified file.

    :param path: the file to write the labels to
    :type path:
    :param labels:
    :param logger:
    :return:
    """
    if logger is not None:
        logger.info("Writing labels CSV file: %s" % path)

    rows = [["Index", "Label"]]
    for key in labels:
        rows.append([labels[key], key])
    with open(path, "w") as fp:
        writer = csv.writer(fp)
        writer.writerows(rows)


def load_function(function: str) -> Callable:
    """
    Parses the function definition and returns the function.
    The default format is "module_name:function_name".
    Raises exceptions if wrong format, missing or not an actual function.

    :param function: the function definition to parse
    :type function: str
    :return: the parsed function
    """
    if ":" not in function:
        raise Exception("Expected format 'module_name:function_name' but got: %s" % function)
    else:
        module_name, func_name = function.split(":")

    try:
        module = importlib.import_module(module_name)
    except:
        raise Exception("Failed to import class lister module: %s" % module_name)

    if hasattr(module, func_name):
        func = getattr(module, func_name)
        if inspect.isfunction(func):
            return func
        else:
            raise Exception("Not an actual function: %s" % function)
    else:
        raise Exception("Function '%s' not found in module '%s'!" % (func_name, module_name))


def pad_image(img: Union[Image.Image, np.ndarray], pad_width: Optional[int] = None, pad_height: Optional[int] = None) -> Image:
    """
    Pads the image/layer if necessary (on the right/bottom).

    :param img: the image to pad
    :type img: Image.Image/np.ndarray
    :param pad_width: the width to pad to, return as is if None
    :type pad_width: int
    :param pad_height: the height to pad to, return as is if None
    :type pad_height: int
    :return: the (potentially) padded image
    :rtype: Image.Image/np.ndarray
    """
    result = img
    if isinstance(img, Image.Image):
        width, height = img.size
    else:
        height = img.shape[0]
        width = img.shape[1]
    pad = False

    if (pad_width is not None) and (pad_height is not None):
        pad = (width != pad_width) or (height != pad_height)
    elif pad_width is not None:
        pad = width != pad_width
        pad_height = height
    elif pad_height is not None:
        pad = height != pad_height
        pad_width = width

    if pad:
        if isinstance(img, Image.Image):
            result = Image.new(img.mode, (pad_width, pad_height))
            result.paste(img)
        else:
            result = np.zeros((pad_height, pad_width), dtype=img.dtype)
            result[0:height, 0:width] = img

    return result


def crop_image(img: Union[Image.Image, np.ndarray], crop_width: Optional[int] = None, crop_height: Optional[int] = None) -> Image:
    """
    Crops the image/layer if necessary (removes on the right/bottom).

    :param img: the image to pad
    :type img: Image.Image/np.ndarray
    :param crop_width: the width to crop to, return as is if None
    :type crop_width: int
    :param crop_height: the height to crop to, return as is if None
    :type crop_height: int
    :return: the (potentially) cropped image
    :rtype: Image.Image/np.ndarray
    """
    result = img
    if isinstance(img, Image.Image):
        width, height = img.size
    else:
        height = img.shape[0]
        width = img.shape[1]
    crop = False

    if (crop_width is not None) and (crop_height is not None):
        crop = (width != crop_width) or (height != crop_height)
    elif crop_width is not None:
        crop = width != crop_width
        crop_height = height
    elif crop_height is not None:
        crop = height != crop_height
        crop_width = width

    if crop:
        if isinstance(img, Image.Image):
            result = img.crop((0, 0, crop_width, crop_height))
        else:
            result = img[0:crop_height, 0:crop_width]

    return result


def safe_deepcopy(obj: Optional[Any]) -> Optional[Any]:
    """
    Creates a deep copy of the object. Skips None objects.

    :param obj: the object to copy, can be None
    :return: the copy or None
    """
    if obj is None:
        return None
    else:
        return copy.deepcopy(obj)



import argparse
from typing import List, Iterable, Union

from wai.logging import LOGGING_WARNING
from wai.common.geometry import NormalizedPoint, NormalizedPolygon
from wai.common.adams.imaging.locateobjects import NormalizedLocatedObjects, NormalizedLocatedObject
from seppl.placeholders import PlaceholderSupporter, placeholder_list
from seppl.io import locate_files
from idc.api import ObjectDetectionData, locate_image, Reader, load_labels


class YoloObjectDetectionReader(Reader, PlaceholderSupporter):

    def __init__(self, source: Union[str, List[str]] = None, source_list: Union[str, List[str]] = None,
                 image_path_rel: str = None, use_polygon_format: bool = False, labels: str = None,
                 resume_from: str = None, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param source: the filename(s)
        :param source_list: the file(s) with filename(s)
        :param image_path_rel: the relative path from the annotations to the images
        :type image_path_rel: str
        :param use_polygon_format: whether to use the polygon format
        :type use_polygon_format: bool
        :param labels: the text file with the comma-separated list of labels
        :type labels: str
        :param resume_from: the file to resume from (glob)
        :type resume_from: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.source = source
        self.source_list = source_list
        self.image_path_rel = image_path_rel
        self.use_polygon_format = use_polygon_format
        self.labels = labels
        self.resume_from = resume_from
        self._label_mapping = None
        self._inputs = None
        self._current_input = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "from-yolo-od"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Loads the bounding box and/or polygon definitions from the associated .txt file in YOLO format."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--input", type=str, help="Path to the text file(s) to read; glob syntax is supported; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("-I", "--input_list", type=str, help="Path to the text file(s) listing the text files to use; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("--resume_from", type=str, help="Glob expression matching the file to resume from, e.g., '*/012345.txt'", required=False)
        parser.add_argument("--image_path_rel", metavar="PATH", type=str, default=None, help="The relative path from the annotations to the images directory", required=False)
        parser.add_argument("-p", "--use_polygon_format", action="store_true", help="Whether to read the annotations in polygon format rather than bbox format", required=False)
        parser.add_argument("--labels", metavar="FILE", type=str, default=None, help="The text file with the comma-separated list of labels", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.source = ns.input
        self.source_list = ns.input_list
        self.image_path_rel = ns.image_path_rel
        self.use_polygon_format = ns.use_polygon_format
        self.labels = ns.labels
        self.resume_from = ns.resume_from

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.image_path_rel is None:
            self.image_path_rel = "../images"
        if self.labels is None:
            raise Exception("No labels file defined!")
        self._inputs = locate_files(self.source, input_lists=self.source_list, fail_if_empty=True, default_glob="*.txt", resume_from=self.resume_from)
        _, self._label_mapping = load_labels(self.labels, logger=self.logger())

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        self._current_input = self._inputs.pop(0)
        self.session.current_input = self._current_input
        self.logger().info("Reading from: " + str(self.session.current_input))
        with open(self.session.current_input, "r") as fp:
            lines = fp.readlines()
            lines = [x.strip() for x in lines]

        annotations = NormalizedLocatedObjects()
        for line in lines:
            if len(line) == 0:
                continue
            parts = line.split(" ")
            if len(parts) < 3:
                continue
            label = int(parts[0])
            meta = {"type": self._label_mapping[label]}
            parts = parts[1:]
            if self.use_polygon_format:
                if len(parts) % 2 != 0:
                    self.logger().warning("Cannot process uneven number of coordinates in polygon format: %s" % line)
                    continue
                points = []
                minx = 1.0
                maxx = 0.0
                miny = 1.0
                maxy = 0.0
                for i in range(len(parts) // 2):
                    x = float(parts[i*2])
                    y = float(parts[i*2+1])
                    point = NormalizedPoint(x, y)
                    points.append(point)
                    minx = min(minx, x)
                    maxx = max(maxx, x)
                    miny = min(miny, y)
                    maxy = max(maxy, y)
                poly = NormalizedPolygon(*points)
                obj = NormalizedLocatedObject(minx, miny, (maxx - minx), (maxy - miny), **meta)
                obj.set_polygon(poly)
                annotations.append(obj)
            else:
                if len(parts) != 4:
                    self.logger().warning("BBox format requires 4 values (left, top, width, height), but got %d instead: %s" % (len(parts), line))
                    continue
                x_center = float(parts[0])
                y_center = float(parts[1])
                width = float(parts[2])
                height = float(parts[3])
                obj = NormalizedLocatedObject(x_center - width / 2, y_center - height / 2, width, height, **meta)
                annotations.append(obj)

        image = locate_image(self._current_input, rel_path=self.image_path_rel)
        if image is None:
            self.logger().warning("No associated image found: %s" % self._current_input)
            yield None

        yield ObjectDetectionData(source=image, annotation=annotations)

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return len(self._inputs) == 0



import argparse
import glob
import os.path
from typing import List, Iterable, Union

import numpy as np
from PIL import Image, ImageOps
from seppl.placeholders import PlaceholderSupporter, placeholder_list
from seppl.io import locate_files
from wai.logging import LOGGING_WARNING

from idc.api import ImageSegmentationData, ImageSegmentationAnnotations
from idc.api import Reader


class LayerSegmentsImageSegmentationReader(Reader, PlaceholderSupporter):

    def __init__(self, source: Union[str, List[str]] = None, source_list: Union[str, List[str]] = None,
                 label_separator: str = "-", labels: List[str] = None,
                 lenient: bool = False, invert: bool = False, resume_from: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param source: the filename(s)
        :param source_list: the file(s) with filename(s)
        :param label_separator: the separator between name and label for the masks
        :type label_separator: str
        :param labels: the list of labels
        :type labels: list
        :param lenient: whether to quietly convert non-binary images with just two unique colors into binary ones or raise an exception
        :type lenient: bool
        :param invert: whether to invert the binary images (b/w <-> w/b)
        :type invert: bool
        :param resume_from: the file to resume from (glob)
        :type resume_from: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.source = source
        self.source_list = source_list
        self.label_separator = label_separator
        self.labels = labels
        self.lenient = lenient
        self.invert = invert
        self.resume_from = resume_from
        self._label_mapping = None
        self._inputs = None
        self._current_input = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "from-layer-segments-is"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Loads the annotations from associated mask PNG image files, with one binary mask per layer/label."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--input", type=str, help="Path to the JPG file(s) to read; glob syntax is supported; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("-I", "--input_list", type=str, help="Path to the text file(s) listing the JPG files to use; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("--resume_from", type=str, help="Glob expression matching the file to resume from, e.g., '*/012345.jpg'", required=False)
        parser.add_argument("--labels", metavar="LABEL", type=str, default=None, help="The labels that the indices represent.", nargs="+")
        parser.add_argument("--label_separator", type=str, help="The separator between name and label used by the mask images.", required=False, default="-")
        parser.add_argument("--lenient", action="store_true", help="Will convert non-binary masks with just two unique color values quietly to binary without raising an exception.")
        parser.add_argument("--invert", action="store_true", help="Will invert the binary images (b/w <-> w/b).")
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.source = ns.input
        self.source_list = ns.input_list
        self.labels = ns.labels
        self.label_separator = ns.label_separator
        self.lenient = ns.lenient
        self.invert = ns.invert
        self.resume_from = ns.resume_from

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ImageSegmentationData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.labels is None:
            raise Exception("No labels defined!")
        self._inputs = locate_files(self.source, input_lists=self.source_list, fail_if_empty=True, default_glob="*.jpg", resume_from=self.resume_from)

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        self._current_input = self._inputs.pop(0)
        self.session.current_input = self._current_input

        # associated layers?
        prefix = os.path.splitext(self.session.current_input)[0]
        anns = glob.glob(prefix + self.label_separator + "*.png")
        if len(anns) == 0:
            self.logger().warning("No associated layers found for: %s" % self.session.current_input)
            self._current_input = None
            yield None

        # read annotations
        prefix_short = os.path.basename(prefix)
        layers = dict()
        for ann in anns:
            self.logger().info("Reading from: " + str(ann))
            ann_short = os.path.splitext(os.path.basename(ann))[0]
            label = ann_short[len(prefix_short + self.label_separator):]
            if label in self.labels:
                img = Image.open(ann)
                if img.mode != "1":
                    arr = np.asarray(img).astype(np.uint8)
                    unique = np.unique(arr)
                    if self.lenient and (len(unique) == 2):
                        for i in range(2):
                            arr = np.where(arr == unique[i], i*255, arr)
                        img = Image.fromarray(arr, "L").convert("1")
                    else:
                        raise Exception("Layer mask is not binary: %s! Try using --lenient flag to attempt loading file(s)." % ann)
                if self.invert:
                    img = ImageOps.invert(img)
                arr = np.asarray(img).astype(np.uint8)
                arr = np.where(arr == 1, 255, arr).astype(np.uint8)
                layers[label] = arr
            else:
                self.logger().warning("Skipping unknown label: %s" % label)

        annotations = ImageSegmentationAnnotations(self.labels, layers)
        yield ImageSegmentationData(source=self.session.current_input, annotation=annotations)

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return len(self._inputs) == 0



import argparse
from typing import List

import numpy as np
from PIL import Image
from seppl import AnyData, AliasSupporter
from seppl.io import Filter
from wai.logging import LOGGING_WARNING

from idc.api import ImageData, flatten_list, make_list, array_to_image, safe_deepcopy

CONVERSION_BT601 = "BT.601"
CONVERSION_BT709 = "BT.709"
CONVERSION_BT2020 = "BT.2020"
CONVERSIONS = [
    CONVERSION_BT601,
    CONVERSION_BT709,
    CONVERSION_BT2020,
]
CONVERSION_PARAMETERS = {
    CONVERSION_BT601: [0.2989, 0.5870, 0.1140],
    CONVERSION_BT709: [0.2126, 0.7152, 0.0722],
    CONVERSION_BT2020: [0.2627, 0.678, 0.0593],
}
CONVERSION_INFO = {
    CONVERSION_BT601: "https://en.wikipedia.org/wiki/Rec._601",
    CONVERSION_BT709: "https://en.wikipedia.org/wiki/Rec._709",
    CONVERSION_BT2020: "https://en.wikipedia.org/wiki/Rec._2020",
}


class RGBToGrayscale(Filter, AliasSupporter):
    """
    Turns RGB images into grayscale ones.
    """

    def __init__(self, conversion: str = CONVERSION_BT601,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param conversion: the conversion to use, BT601 by default
        :type conversion: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.conversion = conversion

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "rgb-to-grayscale"

    def aliases(self) -> List[str]:
        """
        Returns the aliases under which the plugin is known under/available as well.

        :return: the aliases
        :rtype: list
        """
        return ["rgb-to-greyscale"]

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Turns RGB images into grayscale ones.\n" \
            + "\n".join(["%s: %s" % (x, CONVERSION_INFO[x]) for x in CONVERSIONS])

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ImageData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-c", "--conversion", choices=CONVERSIONS, help="The conversion to apply.", required=False, default=CONVERSION_BT601)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.conversion = ns.conversion

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.conversion is None:
            self.conversion = CONVERSION_BT601
        if self.conversion not in CONVERSIONS:
            raise Exception("Unknown conversion (available: %s): %s" % ("|".join(CONVERSIONS), self.conversion))

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []
        for rgb_item in make_list(data):
            rgb_array = np.array(rgb_item.image)
            gray_array = np.dot(rgb_array[..., :3], CONVERSION_PARAMETERS[self.conversion])
            gray_img = Image.fromarray(np.uint8(gray_array), mode='L')
            gray_item = type(rgb_item)(source=None, image_name=rgb_item.image_name,
                                       data=array_to_image(gray_array, rgb_item.image_format),
                                       image=gray_img, image_format=rgb_item.image_format,
                                       metadata=safe_deepcopy(rgb_item.get_metadata()),
                                       annotation=safe_deepcopy(rgb_item.annotation))
            result.append(gray_item)

        return flatten_list(result)



import argparse
import logging
import os
import traceback

from typing import Dict, List, Optional

from seppl import ClassListerRegistry, Plugin, get_class_name

# environment variable with comma-separated list of class listers to use
ENV_IDC_CLASS_LISTERS = "IDC_CLASS_LISTERS"

# environment variable with comma-separated list of class listers to exclude
ENV_IDC_CLASS_LISTERS_EXCL = "IDC_CLASS_LISTERS_EXCL"

# the default class listers to inspect (for development)
# can be overridden with IDC_CLASS_LISTERS environment variable
DEFAULT_IDC_CLASS_LISTERS = [
    "idc.class_lister",
]

REGISTRY = ClassListerRegistry(default_class_listers=DEFAULT_IDC_CLASS_LISTERS,
                               env_class_listers=ENV_IDC_CLASS_LISTERS,
                               env_excluded_class_listers=ENV_IDC_CLASS_LISTERS_EXCL)

IMG_REGISTRY = "idc-registry"

_logger = None


LIST_PLUGINS = "plugins"
LIST_PIPELINE = "pipeline"
LIST_READERS = "readers"
LIST_FILTERS = "filters"
LIST_WRITERS = "writers"
LIST_GENERATORS = "generators"
LIST_CUSTOM_CLASS_LISTERS = "custom-class-listers"
LIST_ENV_CLASS_LISTERS = "env-class-listers"
LIST_TYPES = [
    LIST_PLUGINS,
    LIST_PIPELINE,
    LIST_CUSTOM_CLASS_LISTERS,
    LIST_ENV_CLASS_LISTERS,
    LIST_READERS,
    LIST_FILTERS,
    LIST_WRITERS,
    LIST_GENERATORS,
]


def logger() -> logging.Logger:
    """
    Returns the logger instance to use, initializes it if necessary.

    :return: the logger instance
    :rtype: logging.Logger
    """
    global _logger
    if _logger is None:
        _logger = logging.getLogger(IMG_REGISTRY)
    return _logger


def available_readers() -> Dict[str, Plugin]:
    """
    Returns all available readers.

    :return: the dict of reader objects
    :rtype: dict
    """
    return REGISTRY.plugins("seppl.io.Reader", fail_if_empty=False)


def available_writers() -> Dict[str, Plugin]:
    """
    Returns all available writers.

    :return: the dict of writer objects
    :rtype: dict
    """
    return REGISTRY.plugins("seppl.io.Writer", fail_if_empty=False)


def available_filters() -> Dict[str, Plugin]:
    """
    Returns all available filters.

    :return: the dict of filter objects
    :rtype: dict
    """
    return REGISTRY.plugins("seppl.io.Filter", fail_if_empty=False)


def available_pipeline_plugins() -> Dict[str, Plugin]:
    """
    Returns all available plugins for building pipelines.

    :return: the dict of plugin objects
    :rtype: dict
    """
    result = dict()
    result.update(available_readers())
    result.update(available_filters())
    result.update(available_writers())
    return result


def available_generators() -> Dict[str, Plugin]:
    """
    Returns all available generators.

    :return: the dict of generator objects
    :rtype: dict
    """
    return REGISTRY.plugins("idc.api.Generator", fail_if_empty=False)


def available_plugins() -> Dict[str, Plugin]:
    """
    Returns all available plugins (pipeline and generators).

    :return: the dict of plugin objects
    :rtype: dict
    """
    result = dict()
    result.update(available_readers())
    result.update(available_filters())
    result.update(available_writers())
    result.update(available_generators())
    return result


def register_plugins(custom_class_listers: List[str] = None, excluded_class_listers: List[str] = None):
    """
    Registers all plugins.

    :param custom_class_listers: the list of custom class listers to use instead of env variable or default class listers
    :type custom_class_listers: list
    :param excluded_class_listers: the list of class listers to exclude
    :type excluded_class_listers: list
    """
    REGISTRY.custom_class_listers = custom_class_listers
    REGISTRY.excluded_class_listers = excluded_class_listers
    available_plugins()


def _list(list_type: str, custom_class_listers: Optional[List[str]] = None, excluded_class_listers: Optional[List[str]] = None):
    """
    Lists various things on stdout.

    :param custom_class_listers: the list of custom class listers to use instead of env variable or default class listers
    :type custom_class_listers: list
    :param excluded_class_listers: the list of class listers to exclude
    :type excluded_class_listers: list
    :param list_type: the type of list to generate
    :type list_type: str
    """
    register_plugins(custom_class_listers=custom_class_listers, excluded_class_listers=excluded_class_listers)

    if list_type in [LIST_PLUGINS, LIST_PIPELINE, LIST_READERS, LIST_FILTERS, LIST_WRITERS, LIST_GENERATORS]:
        if list_type == LIST_PLUGINS:
            plugins = available_plugins()
        elif list_type == LIST_PIPELINE:
            plugins = available_pipeline_plugins()
        elif list_type == LIST_READERS:
            plugins = available_readers()
        elif list_type == LIST_FILTERS:
            plugins = available_filters()
        elif list_type == LIST_WRITERS:
            plugins = available_readers()
        elif list_type == LIST_GENERATORS:
            plugins = available_generators()
        else:
            raise Exception("Unhandled type: %s" % list_type)
        print("name: class")
        for name in plugins:
            print("%s: %s" % (name, get_class_name(plugins[name])))
    elif list_type == LIST_CUSTOM_CLASS_LISTERS:
        class_listers = REGISTRY.custom_class_listers
        print("custom class listers:")
        if class_listers is None:
            print("-none")
        else:
            for m in class_listers:
                print(m)
    elif list_type == LIST_ENV_CLASS_LISTERS:
        print("env class listers:")
        if REGISTRY.env_class_listers is None:
            print("-none-")
        else:
            class_listers = os.getenv(REGISTRY.env_class_listers)
            if class_listers is None:
                print("-none listed in env var %s-" % REGISTRY.env_class_listers)
            else:
                class_listers = class_listers.split(",")
                for m in class_listers:
                    print(m)


def main(args=None):
    """
    The main method for parsing command-line arguments.

    :param args: the commandline arguments, uses sys.argv if not supplied
    :type args: list
    """
    parser = argparse.ArgumentParser(
        description="For inspecting/querying the registry.",
        prog=IMG_REGISTRY,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-c", "--custom_class_listers", type=str, default=None, help="The comma-separated list of custom class listers to use.", required=False)
    parser.add_argument("-e", "--excluded_class_listers", type=str, default=None, help="The comma-separated list of class listers to exclude.", required=False)
    parser.add_argument("-l", "--list", choices=LIST_TYPES, default=None, help="For outputting various lists on stdout.", required=False)
    parsed = parser.parse_args(args=args)

    custom_class_listers = None
    if parsed.custom_class_listers is not None:
        custom_class_listers = parsed.custom_class_listers.split(",")

    if parsed.list is not None:
        _list(parsed.list, custom_class_listers=custom_class_listers)


def sys_main() -> int:
    """
    Runs the main function using the system cli arguments, and
    returns a system error code.

    :return: 0 for success, 1 for failure.
    """
    try:
        main()
        return 0
    except Exception:
        traceback.print_exc()
        return 1


if __name__ == '__main__':
    main()



import os

from seppl import OutputProducer, InputConsumer, classes_to_str, get_aliases, has_aliases
from seppl.placeholders import PlaceholderSupporter, placeholder_help

from idc.registry import available_plugins


HELP_FORMAT_TEXT = "text"
HELP_FORMAT_MARKDOWN = "markdown"
HELP_FORMATS = [
    HELP_FORMAT_TEXT,
    HELP_FORMAT_MARKDOWN,
]


def generate_plugin_usage(plugin_name: str, help_format: str = HELP_FORMAT_TEXT, heading_level: int = 1,
                          output_path: str = None):
    """
    Generates the usage help screen for the specified plugin.

    :param plugin_name: the plugin to generate the usage for (name used on command-line)
    :type plugin_name: str
    :param help_format: the format to use for the output
    :type help_format: str
    :param heading_level: the level to use for the heading (markdown)
    :type heading_level: int
    :param output_path: the directory (automatically generates output name from plugin name and output format) or file to store the generated help in, uses stdout if None
    :type output_path: str
    """
    if help_format not in HELP_FORMATS:
        raise Exception("Unhandled help format: %s" % help_format)

    plugin = available_plugins()[plugin_name]

    result = ""
    if help_format == HELP_FORMAT_TEXT:
        suffix = ".txt"
        result += "\n" + plugin_name + "\n" + "=" * len(plugin_name) + "\n"
        if isinstance(plugin, InputConsumer):
            result += "accepts: " + classes_to_str(plugin.accepts()) + "\n"
        if isinstance(plugin, OutputProducer):
            result += "generates: " + classes_to_str(plugin.generates()) + "\n"
        if has_aliases(plugin):
            result += "alias(es): " + ", ".join(get_aliases(plugin)) + "\n"
        result = result.strip()
        result += "\n\n"
        result += plugin.format_help() + "\n"
        if isinstance(plugin, PlaceholderSupporter):
            result += "\n" + placeholder_help(markdown=False, obj=plugin) + "\n"
    elif help_format == HELP_FORMAT_MARKDOWN:
        suffix = ".md"
        result += "#"*heading_level + " " + plugin_name + "\n"
        result += "\n"
        if isinstance(plugin, InputConsumer):
            result += "* accepts: " + classes_to_str(plugin.accepts()) + "\n"
        if isinstance(plugin, OutputProducer):
            result += "* generates: " + classes_to_str(plugin.generates()) + "\n"
        if has_aliases(plugin):
            result += "* alias(es): " + ", ".join(get_aliases(plugin)) + "\n"
        result = result.strip()
        result += "\n\n"
        result += plugin.description() + "\n"
        result += "\n"
        result += "```\n"
        result += plugin.format_help()
        result += "```\n"
        if isinstance(plugin, PlaceholderSupporter):
            result += "\n" + placeholder_help(markdown=True, obj=plugin) + "\n"
    else:
        raise Exception("Unhandled help format: %s" % help_format)

    if output_path is None:
        print(result)
    else:
        if os.path.isdir(output_path):
            output_file = os.path.join(output_path, plugin.name() + suffix)
        else:
            output_file = output_path
        with open(output_file, "w") as fp:
            fp.write(result)



import argparse
import os
from typing import List
from xml.etree.ElementTree import Element, ElementTree

from wai.logging import LOGGING_WARNING
from idc.api import ObjectDetectionData, SplittableStreamWriter, get_object_label, make_list, AnnotationsOnlyWriter, add_annotations_only_param
from seppl.placeholders import placeholder_list, InputBasedPlaceholderSupporter


def append_element(to: Element, tag: str, value):
    """
    Helper method to append a simple element to a parent element.

    :param to: The element to append the new element to.
    :param tag: The tag-type of the new element.
    :param value: The text content of the new element.
    """
    element = Element(tag)
    element.text = str(value)
    to.append(element)


class VOCObjectDetectionWriter(SplittableStreamWriter, AnnotationsOnlyWriter, InputBasedPlaceholderSupporter):

    def __init__(self, output_dir: str = None, annotations_only: bool = None,
                 split_names: List[str] = None, split_ratios: List[int] = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the writer.

        :param output_dir: the output directory to save the image/report in
        :type output_dir: str
        :param annotations_only: whether to output only the annotations and not the images
        :type annotations_only: bool
        :param split_names: the names of the splits, no splitting if None
        :type split_names: list
        :param split_ratios: the integer ratios of the splits (must sum up to 100)
        :type split_ratios: list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(split_names=split_names, split_ratios=split_ratios, logger_name=logger_name, logging_level=logging_level)
        self.output_dir = output_dir
        self.annotations_only = annotations_only

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "to-voc-od"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Saves the bounding box definitions in PASCAL VOC .xml format."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-o", "--output", type=str, help="The directory to store the images/.xml files in. Any defined splits get added beneath there. " + placeholder_list(obj=self), required=True)
        add_annotations_only_param(parser)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.output_dir = ns.output
        self.annotations_only = ns.annotations_only

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.annotations_only is None:
            self.annotations_only = False

    def _append_element(self, to: Element, tag: str, value):
        """
        Helper method to append a simple element to a parent element.

        :param to: The element to append the new element to.
        :param tag: The tag-type of the new element.
        :param value: The text content of the new element.
        """
        element = Element(tag)
        element.text = str(value)
        to.append(element)

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        for item in make_list(data):
            sub_dir = self.session.expand_placeholders(self.output_dir)
            if self.splitter is not None:
                split = self.splitter.next(item=item.image_name)
                sub_dir = os.path.join(sub_dir, split)
            if not os.path.exists(sub_dir):
                self.logger().info("Creating dir: %s" % sub_dir)
                os.makedirs(sub_dir)

            absolute = None
            if item.has_annotation():
                absolute = item.get_absolute()

            # image
            path = sub_dir
            os.makedirs(path, exist_ok=True)
            path = os.path.join(path, item.image_name)
            if not self.annotations_only:
                self.logger().info("Writing image to: %s" % path)
                item.save_image(path)

            # annotations
            if absolute is not None:
                root = Element("annotation")

                append_element(root, "folder", os.path.dirname(path))
                append_element(root, "filename", os.path.basename(path))

                size = Element("size")
                root.append(size)
                append_element(size, "width", item.image_width)
                append_element(size, "height", item.image_height)
                append_element(size, "depth", 3)

                for lobj in absolute:
                    obj = Element("object")
                    root.append(obj)

                    append_element(obj, "name", get_object_label(lobj))
                    append_element(obj, "pose", "Unspecified")
                    append_element(obj, "truncated", 0)
                    append_element(obj, "difficult", 0)

                    bndbox = Element("bndbox")
                    obj.append(bndbox)
                    append_element(bndbox, "xmin", lobj.x)
                    append_element(bndbox, "ymin", lobj.y)
                    append_element(bndbox, "xmax", lobj.x + lobj.width - 1)
                    append_element(bndbox, "ymax", lobj.y + lobj.height - 1)

                path = sub_dir
                os.makedirs(path, exist_ok=True)
                path = os.path.join(path, item.image_name)
                path = os.path.splitext(path)[0] + ".xml"
                self.logger().info("Writing annotations to: %s" % path)
                tree = ElementTree(root)
                tree.write(path, "utf-8", short_empty_elements=False)



import argparse
import os
from collections import OrderedDict
from typing import List

from wai.logging import LOGGING_WARNING
from idc.api import ObjectDetectionData, SplittableStreamWriter, save_labels, save_labels_csv, make_list, AnnotationsOnlyWriter, add_annotations_only_param
from seppl.placeholders import placeholder_list, InputBasedPlaceholderSupporter


class YoloObjectDetectionWriter(SplittableStreamWriter, AnnotationsOnlyWriter, InputBasedPlaceholderSupporter):

    def __init__(self, output_dir: str = None,
                 image_subdir: str = None, labels_subdir: str = None, categories: List[str] = None,
                 use_polygon_format: bool = False, labels: str = None, labels_csv: str = None,
                 annotations_only: bool = None,
                 split_names: List[str] = None, split_ratios: List[int] = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the writer.

        :param output_dir: the output directory to save the image/report in
        :type output_dir: str
        :param image_subdir: the sub-dir to use for the images
        :type image_subdir: str
        :param labels_subdir: the sub-dir to use for the labels
        :type labels_subdir: str
        :param categories: the predefined categories to use
        :type categories: str
        :param use_polygon_format: whether to use the polygon format
        :type use_polygon_format: bool
        :param labels: the text file with the comma-separated list of labels
        :type labels: str
        :param labels_csv: the CSV file to write the label mapping to (index and label)
        :type labels_csv: str
        :param annotations_only: whether to output only the annotations and not the images
        :type annotations_only: bool
        :param split_names: the names of the splits, no splitting if None
        :type split_names: list
        :param split_ratios: the integer ratios of the splits (must sum up to 100)
        :type split_ratios: list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(split_names=split_names, split_ratios=split_ratios, logger_name=logger_name, logging_level=logging_level)
        self.output_dir = output_dir
        self.image_subdir = image_subdir
        self.labels_subdir = labels_subdir
        self.categories = categories
        self.use_polygon_format = use_polygon_format
        self.labels = labels
        self.labels_csv = labels_csv
        self.annotations_only = annotations_only
        self._label_mapping = None  # label -> index
        self._output_dirs = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "to-yolo-od"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Saves the bounding box/polygon definitions in YOLO .txt format. By default, places images in the 'images' subdir and the annotations in 'labels'."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-o", "--output", type=str, help="The directory to store the images/.txt files in. Any defined splits get added beneath there. " + placeholder_list(obj=self), required=True)
        parser.add_argument("--image_subdir", metavar="DIR", type=str, default=None, help="The name of the sub-dir to use for storing the images in.", required=False)
        parser.add_argument("--labels_subdir", metavar="DIR", type=str, default=None, help="The name of the sub-dir to use for storing the annotations in.", required=False)
        parser.add_argument("-p", "--use_polygon_format", action="store_true", help="Whether to write the annotations in polygon format rather than bbox format", required=False)
        parser.add_argument("--categories", type=str, help="The predefined order of categories.", required=False, nargs="*")
        parser.add_argument("--labels", type=str, default=None, help="The text file (no path) with the comma-separated list of labels", required=False)
        parser.add_argument("--labels_csv", type=str, default=None, help="The CSV file (no path) to write the label mapping to (index and label)", required=False)
        add_annotations_only_param(parser)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.output_dir = ns.output
        self.image_subdir = ns.image_subdir
        self.labels_subdir = ns.labels_subdir
        self.categories = ns.categories
        self.use_polygon_format = ns.use_polygon_format
        self.labels = ns.labels
        self.labels_csv = ns.labels_csv
        self.annotations_only = ns.annotations_only

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.labels is None:
            raise Exception("No output file for labels provided!")
        if self.image_subdir is None:
            self.image_subdir = "images"
        if self.labels_subdir is None:
            self.labels_subdir = "labels"
        if self.annotations_only is None:
            self.annotations_only = False
        self._label_mapping = OrderedDict()
        if self.categories is not None:
            for i, category in enumerate(self.categories):
                self._label_mapping[category] = i
        self._output_dirs = []

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        for item in make_list(data):
            sub_dir = self.session.expand_placeholders(self.output_dir)
            if self.splitter is not None:
                split = self.splitter.next(item=item.image_name)
                sub_dir = os.path.join(sub_dir, split)
            if not os.path.exists(sub_dir):
                self.logger().info("Creating dir: %s" % sub_dir)
                os.makedirs(sub_dir)
            if sub_dir not in self._output_dirs:
                self._output_dirs.append(sub_dir)

            normalized = None
            if item.has_annotation():
                normalized = item.get_normalized()

            # image
            path = sub_dir
            if len(self.image_subdir) > 0:
                path = os.path.join(path, self.image_subdir)
            os.makedirs(path, exist_ok=True)
            path = os.path.join(path, item.image_name)
            if not self.annotations_only:
                self.logger().info("Writing image to: %s" % path)
                item.save_image(path)

            # annotations
            if normalized is not None:
                lines = []
                for obj in normalized:
                    # determine index
                    index = 0
                    if (obj.metadata is not None) and ("type" in obj.metadata):
                        label = obj.metadata["type"]
                        if label not in self._label_mapping:
                            self._label_mapping[label] = len(self._label_mapping)
                        index = self._label_mapping[label]
                    # determine coordinates
                    if self.use_polygon_format:
                        values = []
                        for x, y in zip(obj.get_polygon_x(), obj.get_polygon_y()):
                            values.append(x)
                            values.append(y)
                    else:
                        values = [obj.x + obj.width / 2, obj.y + obj.height / 2, obj.width, obj.height]
                    values = ["%.6f" % value for value in values]
                    line = str(index) + " " + " ".join(values)
                    lines.append(line)

                path = sub_dir
                if len(self.labels_subdir) > 0:
                    path = os.path.join(path, self.labels_subdir)
                os.makedirs(path, exist_ok=True)
                path = os.path.join(path, item.image_name)
                path = os.path.splitext(path)[0] + ".txt"
                self.logger().info("Writing annotations to: %s" % path)
                with open(path, "w") as fp:
                    for line in lines:
                        fp.write(line)
                        fp.write("\n")

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        super().finalize()

        for sub_dir in self._output_dirs:
            # labels
            save_labels(os.path.join(sub_dir, self.labels), self._label_mapping.keys(), logger=self.logger())

            # labels csv
            if self.labels_csv is not None:
                save_labels_csv(os.path.join(sub_dir, self.labels_csv), self._label_mapping, logger=self.logger())



from ._bluechannel import BlueChannelImageSegmentationReader
from ._grayscale import GrayscaleImageSegmentationReader
from ._indexedpng import IndexedPngImageSegmentationReader
from ._layersegments import LayerSegmentsImageSegmentationReader



from ._adams import AdamsObjectDetectionReader
from ._coco import COCOObjectDetectionReader
from ._opex import OPEXObjectDetectionReader
from ._roi import ROIObjectDetectionReader
from ._voc import VOCObjectDetectionReader
from ._yolo import YoloObjectDetectionReader



import logging

import seppl


ENV_IDC_LOGLEVEL = "IDC_LOGLEVEL"
""" environment variable for the global default logging level. """


COMPARISON_LESSTHAN = "lt"
COMPARISON_LESSOREQUAL = "le"
COMPARISON_EQUAL = "eq"
COMPARISON_NOTEQUAL = "ne"
COMPARISON_GREATEROREQUAL = "ge"
COMPARISON_GREATERTHAN = "gt"
COMPARISON_CONTAINS = "contains"
COMPARISON_MATCHES = "matches"

COMPARISONS = [
    COMPARISON_LESSTHAN,
    COMPARISON_LESSOREQUAL,
    COMPARISON_EQUAL,
    COMPARISON_NOTEQUAL,
    COMPARISON_GREATEROREQUAL,
    COMPARISON_GREATERTHAN,
]
COMPARISON_HELP = COMPARISON_LESSTHAN + ": less than, " \
    + COMPARISON_LESSOREQUAL + ": less or equal, " \
    + COMPARISON_EQUAL + ": equal, " \
    + COMPARISON_NOTEQUAL + ": not equal, " \
    + COMPARISON_GREATERTHAN + ": greater than, " \
    + COMPARISON_GREATEROREQUAL + ": greater of equal"

COMPARISONS_EXT = COMPARISONS[:]
COMPARISONS_EXT.append(COMPARISON_CONTAINS)
COMPARISONS_EXT.append(COMPARISON_MATCHES)
COMPARISON_EXT_HELP = COMPARISON_HELP + ", " \
                      + COMPARISON_CONTAINS + ": substring match, " \
                      + COMPARISON_MATCHES + ": regexp match"


class Session(seppl.Session):
    """
    Session object shared among reader, filter(s), writer.
    """
    logger: logging.Logger = logging.getLogger("image-dataset-converter")
    """ the global logger. """



import abc
import argparse
from typing import List, Dict, Optional

from wai.logging import LOGGING_WARNING
from seppl import Plugin, PluginWithLogging, Initializable, split_cmdline, split_args, args_to_objects


class Generator(PluginWithLogging, Initializable, abc.ABC):
    """
    Base class for generators.
    """

    def __init__(self, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the generator.

        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)

    def _check(self) -> Optional[str]:
        """
        Hook method for performing checks.

        :return: None if checks successful, otherwise error message
        :rtype: str
        """
        return None

    def _do_generate(self) -> List[Dict[str, str]]:
        """
        Generates the variables.

        :return: the list of variable dictionaries
        :rtype: list
        """
        raise NotImplementedError()

    def generate(self):
        """
        Generates the variables.

        :return: the list of variable dictionaries
        :rtype: list
        """
        msg = self._check()
        if msg is not None:
            raise Exception(msg)
        return self._do_generate()

    @classmethod
    def parse_generator(cls, cmdline: str) -> 'Generator':
        """
        Parses the commandline and returns the generator plugin.

        :param cmdline: the command-line to parse
        :type cmdline: str
        :return: the generator plugin
        :rtype: Generator
        """
        result = cls.parse_generators(cmdline)
        if len(result) != 1:
            raise Exception("Expected a single generator, but got: %d" % len(result))
        return result[0]

    @classmethod
    def parse_generators(cls, cmdline: str) -> List['Generator']:
        """
        Parses the commandline and returns the list of generator plugins.

        :param cmdline: the command-line to parse
        :type cmdline: str
        :return: the list of generator plugins
        :rtype: list
        """
        from idc.registry import available_generators
        generator_args = split_args(split_cmdline(cmdline), list(available_generators().keys()))
        return args_to_objects(generator_args, available_generators())


class SingleVariableGenerator(Generator):
    """
    Ancestor for generators that output just a single variable.
    """

    def __init__(self, var_name: str = None, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the generator.

        :param var_name: the variable name
        :type var_name: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.var_name = var_name

    def _default_var_name(self) -> str:
        """
        Returns the default variable name.

        :return: the default name
        :rtype: str
        """
        raise NotImplementedError()

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-n", "--var_name", type=str, metavar="NAME", default=self._default_var_name(), help="The name of the variable", required=(self._default_var_name() is None))
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.var_name = ns.var_name

    def _check(self) -> Optional[str]:
        """
        Hook method for performing checks.

        :return: None if checks successful, otherwise error message
        :rtype: str
        """
        result = super()._check()

        if self.var_name is None:
            result = "No variable name provided!"

        return result



import copy
from typing import List, Tuple, Union

from wai.logging import LOGGING_WARNING
from wai.common.geometry import Polygon, Point, NormalizedPolygon, NormalizedPoint
from wai.common.adams.imaging.locateobjects import LocatedObjects, NormalizedLocatedObjects
from seppl.io import Filter
from idc.api import ObjectDetectionData, flatten_list, make_list


class CoerceMask(Filter):
    """
    Coerces the bounds of the annotations to all be polygon-masks.
    Annotations which already have polygons keep theirs, but those
    without are given a rectangular polygon in the shape of their
    bounding box.
    """

    def __init__(self, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "coerce-mask"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Coerces the bounds of the annotations to all be polygon-masks. " \
               "Annotations which already have polygons keep theirs, but those " \
               "without are given a rectangular polygon in the shape of their bounding box."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def _process_annotations(self, ann: Union[LocatedObjects, NormalizedLocatedObjects]) -> Tuple[Union[LocatedObjects, NormalizedLocatedObjects], int]:
        """
        Processes the annotations.

        :param ann: the annotations to process
        :type ann: LocatedObjects
        :return: the tuple of processed annotations and how many were updated
        :rtype: tuple
        """
        normalized = isinstance(ann, NormalizedLocatedObjects)
        if normalized:
            ann_new = LocatedObjects()
        else:
            ann_new = NormalizedLocatedObjects()
        updated = 0

        for obj in ann:
            obj_new = copy.deepcopy(obj)
            if not obj_new.has_polygon():
                updated += 1
                if normalized:
                    points = [Point(obj_new.x, obj_new.y),
                              Point(obj_new.x + obj_new.width - 1, obj_new.y),
                              Point(obj_new.x + obj_new.width - 1, obj_new.y + obj_new.height - 1),
                              Point(obj_new.x, obj_new.y + obj_new.height - 1)]
                    poly = Polygon(*points)
                else:
                    points = [NormalizedPoint(obj_new.x, obj_new.y),
                              NormalizedPoint(obj_new.x + obj_new.width, obj_new.y),
                              NormalizedPoint(obj_new.x + obj_new.width, obj_new.y + obj_new.height),
                              NormalizedPoint(obj_new.x, obj_new.y + obj_new.height)]
                    poly = NormalizedPolygon(*points)
                obj_new.set_polygon(poly)
            ann_new.append(obj_new)

        return ann_new, updated

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            ann, updated = self._process_annotations(item.annotation)
            if updated > 0:
                self.logger().info("Updated %d object(s)" % updated)
                item = item.duplicate(annotation=ann)
            result.append(item)

        return flatten_list(result)



import base64
import io
import logging
from typing import Tuple, Dict, List

import numpy as np
from PIL import Image

from ._data import ImageData


class ImageSegmentationAnnotations:
    """
    Container for image segmentation annotations.
    """

    def __init__(self, labels: List[str] = None, layers: Dict[str, np.ndarray] = None):
        """
        Initializes the container.

        :param labels: the list of labels
        :param layers: the label -> numpy array association, binary (0/255), uint8
        """
        if (labels is not None) and (layers is not None):
            for label in layers:
                if label not in labels:
                    raise Exception("Layer %s is not specified as label!" % label)
        self.labels = labels
        for label in layers:
            if layers[label].dtype != np.uint8:
                raise Exception("Layers must be %s, but got %s for label '%s'!" % (str(np.dtype(np.uint8)), str(layers[label].dtype), label))
        self.layers = layers

    def subset(self, labels: List[str]) -> 'ImageSegmentationAnnotations':
        """
        Returns the subset of annotations based on the supplied labels.

        :param labels: the labels that will make up the subset
        :type labels: list
        :return: the new annotations
        :rtype: ImageSegmentationData
        """
        layers = dict()
        for label in labels:
            if label in self.layers:
                layers[label] = self.layers[label]
        return ImageSegmentationAnnotations(labels=labels, layers=layers)


class ImageSegmentationData(ImageData):
    """
    The annotations are the classification label.
    """

    def __init__(self, source: str = None, image_name: str = None, data: bytes = None,
                 image: Image.Image = None, image_format: str = None, image_size: Tuple[int, int] = None,
                 metadata: Dict = None, annotation: ImageSegmentationAnnotations = None):

        super().__init__(source=source, image_name=image_name, data=data,
                         image=image, image_format=image_format, image_size=image_size,
                         metadata=metadata, annotation=annotation)

    def has_annotation(self) -> bool:
        """
        Checks whether annotations are present.

        :return: True if annotations present
        :rtype: bool
        """
        return (self.annotation is not None) and (len(self.annotation.layers) > 0)

    def _annotation_to_dict(self):
        """
        Turns the annotations into a dictionary.

        :return: the generated dictionary
        :rtype: dict
        """
        result = dict()
        if self.annotation.labels is not None:
            result["labels"] = self.annotation.labels[:]
        result["layers"] = dict()
        for label in self.annotation.layers:
            arr = self.annotation.layers[label]
            ann = Image.fromarray(arr, "L").convert("1")
            buffer = io.BytesIO()
            ann.save(buffer, format=self.image_format)
            result["layers"][label] = base64.encodebytes(buffer.getvalue()).decode("ascii")

        return result

    def new_layer(self, label: str) -> np.ndarray:
        """
        Adds an empty layer with the specified label.

        :param label: the label of the layer
        :type label: str
        :return: the generated layer
        :rtype: np.ndarray
        """
        if self.image_size is None:
            raise Exception("No image dimensions available, cannot create empty layer ''%s!" % label)
        if self.annotation is None:
            self.annotation = ImageSegmentationAnnotations(labels=[], layers={})
        width, height = self.image_size
        layer = np.zeros((height, width), dtype=np.uint8)
        if label not in self.annotation.labels:
            self.annotation.labels.append(label)
        self.annotation.layers[label] = layer
        return layer

    def has_layer(self, label: str) -> bool:
        """
        Checks whether the layer is present.

        :param label: the layer to check
        :type label: str
        :return: true if present (may be empty though)
        :rtype: bool
        """
        if not self.has_annotation():
            return False
        return label in self.annotation.layers


def combine_layers(item: ImageSegmentationData, dtype=np.int32) -> np.ndarray:
    """
    Combines the layers into a single numpy array. The first label gets value 1.

    :param item: the segmentation data to combine
    :type item: ImageSegmentationData
    :param dtype: the data type to use for the array
    :return: the generated array
    :rtype: np.ndarray
    """
    result = np.zeros((item.image_height, item.image_width, 1), dtype=dtype)
    for i, label in enumerate(item.annotation.labels, start=1):
        if label in item.annotation.layers:
            layer = item.annotation.layers[label]
            layer = np.where(layer > 0, i, layer)
            layer = np.expand_dims(layer, axis=-1)
            result += layer
    return result


def split_layers(array: np.ndarray, labels: List[str]) -> ImageSegmentationAnnotations:
    """
    Generates annotations from the annotations array. Assumes the first label having value of 1.

    :param array: the array to turn into annotations
    :type array: np.ndarray
    :param labels: the list of labels to use
    :type labels: list
    :return: the generated annotations
    :rtype: ImageSegmentationAnnotations
    """
    layers = dict()
    unique = set(np.unique(array))
    for i, label in enumerate(labels, start=1):
        if i in unique:
            layer = np.where(array == i, 255, 0)
            layer = np.squeeze(layer)
            layers[label] = layer.astype(np.uint8)
    return ImageSegmentationAnnotations(labels=labels, layers=layers)


def from_indexedpng(img: Image.Image, labels: List[str], label_mapping: Dict[int, str],
                    logger: logging.Logger, background: int = 0) -> ImageSegmentationAnnotations:
    """
    Loads the annotations from the indexed png.

    :param img: the image to turn into annotations
    :type img: Image.Image
    :param labels: the list of labels
    :type labels: list
    :param label_mapping: the mapping of index to label
    :type label_mapping: dict
    :param logger: the (optional) logger for logging messages
    :type logger: logging.Logger
    :param background: the index (0-255) of the background, default 0
    :type background: int
    :return: the generated annotations
    :rtype: ImageSegmentationAnnotations
    """
    arr = np.asarray(img).astype(np.uint8)
    unique = np.unique(arr)
    layers = dict()
    for index in list(unique):
        # skip background
        if index == background:
            continue
        index = int(index)
        if background < index:
            label_index = index - 1
        else:
            label_index = index
        if label_index not in label_mapping:
            msg = "Index not covered by labels, skipping: %d" % index
            if logger is not None:
                logger.warning(msg)
            else:
                print(msg)
            continue
        sub_arr = np.where(arr == index, 255, 0).astype(np.uint8)
        layers[label_mapping[label_index]] = sub_arr
    return ImageSegmentationAnnotations(labels, layers)


def to_indexedpng(width: int, height: int, ann: ImageSegmentationAnnotations, palette_list: List[int], background: int = 0) -> Image.Image:
    """
    Turns the annotations into an indexed image.

    :param width: the width of the image
    :type width: int
    :param height: the height of the image
    :type height: int
    :param ann: the annotations to convert
    :type ann: ImageSegmentationAnnotations
    :param palette_list: the flat list of int RGB values to use for the palette (len = 3 * #labels)
    :type palette_list: list
    :param background: the background index
    :type background: int
    :return: the generated indexed image
    :rtype: Image.Image
    """
    # combine layers
    arr = np.zeros((height, width)).astype(dtype=np.uint8)
    for index, label in enumerate(ann.labels, start=1):
        if label in ann.layers:
            sub_arr = ann.layers[label]
            sub_arr = np.where(sub_arr == 255, index, 0).astype(np.uint8)
            arr += sub_arr
    if background > 0:
        arr = np.where(arr == 0, background, arr)
    result = Image.fromarray(arr, "P")
    result.putpalette(palette_list)
    return result


def from_bluechannel(img: Image.Image, labels: List[str], label_mapping: Dict[int, str],
                     logger: logging.Logger, background: int = 0) -> ImageSegmentationAnnotations:
    """
    Loads the annotations from the blue channel.

    :param img: the image to turn into annotations
    :type img: Image.Image
    :param labels: the list of labels
    :type labels: list
    :param label_mapping: the mapping of index to label
    :type label_mapping: dict
    :param logger: the (optional) logger for logging messages
    :type logger: logging.Logger
    :param background: the index (0-255) of the background, default 0
    :type background: int
    :return: the generated annotations
    :rtype: ImageSegmentationAnnotations
    """
    arr = np.asarray(img).astype(np.uint8)
    arr = arr[:, :, 2]
    unique = np.unique(arr)
    layers = dict()
    for index in list(unique):
        # skip background
        if index == background:
            continue
        index = int(index)
        if background < index:
            label_index = index - 1
        else:
            label_index = index
        if label_index not in label_mapping:
            msg = "Blue channel value not covered by labels, skipping: %d" % index
            if logger is not None:
                logger.warning(msg)
            else:
                print(msg)
            continue
        sub_arr = np.where(arr == index, 255, 0).astype(np.uint8)
        layers[label_mapping[label_index]] = sub_arr
    return ImageSegmentationAnnotations(labels, layers)


def to_bluechannel(width: int, height: int, ann: ImageSegmentationAnnotations, background: int = 0) -> Image.Image:
    """
    Turns the annotations into a RGB image with the layers in the blue channel.

    :param width: the width of the image
    :type width: int
    :param height: the height of the image
    :type height: int
    :param ann: the annotations to convert
    :type ann: ImageSegmentationAnnotations
    :param background: the background index
    :type background: int
    :return: the generated RGB image with the layers in the blue channel
    :rtype: Image.Image
    """
    arr = np.zeros((height, width)).astype(dtype=np.uint8)
    for index, label in enumerate(ann.labels, start=1):
        if label in ann.layers:
            sub_arr = ann.layers[label]
            sub_arr = np.where(sub_arr == 255, index, 0).astype(np.uint8)
            arr += sub_arr
    if background > 0:
        arr = np.where(arr == 0, background, arr)
    blue = np.zeros((*arr.shape, 3), np.uint8)
    blue[:, :, 2] = arr
    result = Image.fromarray(blue, "RGB")
    return result


def from_grayscale(img: Image.Image, labels: List[str], label_mapping: Dict[int, str],
                   logger: logging.Logger, background: int = 0) -> ImageSegmentationAnnotations:
    """
    Loads the annotations from the grayscale image.

    :param img: the image to turn into annotations
    :type img: Image.Image
    :param labels: the list of labels
    :type labels: list
    :param label_mapping: the mapping of index to label
    :type label_mapping: dict
    :param logger: the (optional) logger for logging messages
    :type logger: logging.Logger
    :param background: the index (0-255) of the background, default 0
    :type background: int
    :return: the generated annotations
    :rtype: ImageSegmentationAnnotations
    """
    arr = np.asarray(img).astype(np.uint8)
    unique = np.unique(arr)
    layers = dict()
    for index in list(unique):
        # skip background
        if index == background:
            continue
        index = int(index)
        if background < index:
            label_index = index - 1
        else:
            label_index = index
        if label_index not in label_mapping:
            msg = "Grayscale value not covered by labels, skipping: %d" % index
            if logger is not None:
                logger.warning(msg)
            else:
                print(msg)
            continue
        sub_arr = np.where(arr == index, 255, 0).astype(np.uint8)
        layers[label_mapping[label_index]] = sub_arr
    return ImageSegmentationAnnotations(labels, layers)


def to_grayscale(width: int, height: int, ann: ImageSegmentationAnnotations, background: int = 0) -> Image.Image:
    """
    Turns the annotations into a grayscale image.

    :param width: the width of the image
    :type width: int
    :param height: the height of the image
    :type height: int
    :param ann: the annotations to convert
    :type ann: ImageSegmentationAnnotations
    :param background: the background index
    :type background: int
    :return: the generated grayscale image
    :rtype: Image.Image
    """
    arr = np.zeros((height, width)).astype(dtype=np.uint8)
    for index, label in enumerate(ann.labels, start=1):
        if label in ann.layers:
            sub_arr = ann.layers[label]
            sub_arr = np.where(sub_arr == 255, index, 0).astype(np.uint8)
            arr += sub_arr
    if background > 0:
        arr = np.where(arr == 0, background, arr)
    result = Image.fromarray(arr, "L")
    return result



import argparse
import logging
import traceback

from wai.logging import init_logging, set_logging_level, add_logging_level

from idc.api import Generator
from idc.core import ENV_IDC_LOGLEVEL
from idc.registry import available_generators

TEST_GENERATOR = "idc-test-generator"

_logger = logging.getLogger(TEST_GENERATOR)


def test_generator(generator: str):
    """
    Parses/executes the generator and then outputs the generated variables.

    :param generator: the generator command-line to use for generating variable values
    :type generator: str
    """
    # parse generator
    generator_obj = Generator.parse_generator(generator)

    # apply generator to pipeline template and execute it
    vars_list = generator_obj.generate()
    for vars_ in vars_list:
        print(vars_)


def main(args=None):
    """
    The main method for parsing command-line arguments.

    :param args: the commandline arguments, uses sys.argv if not supplied
    :type args: list
    """
    init_logging(env_var=ENV_IDC_LOGLEVEL)
    generators = sorted(list(available_generators().keys()))
    parser = argparse.ArgumentParser(
        description="Tool for testing generators by outputting the generated variables and their associated values. Available generators: " + ", ".join(generators),
        prog=TEST_GENERATOR,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-g", "--generator", help="The generator plugin to use.", default=None, type=str, required=True)
    add_logging_level(parser)
    parsed = parser.parse_args(args=args)
    set_logging_level(_logger, parsed.logging_level)
    test_generator(parsed.generator)


def sys_main() -> int:
    """
    Runs the main function using the system cli arguments, and
    returns a system error code.

    :return: 0 for success, 1 for failure.
    """
    try:
        main()
        return 0
    except Exception:
        traceback.print_exc()
        return 1


if __name__ == '__main__':
    main()



import seppl.io
from seppl import Initializable


class Reader(seppl.io.Reader, Initializable):
    """
    Ancestor for dataset readers.
    """
    pass


def parse_reader(reader: str) -> Reader:
    """
    Parses the command-line and instantiates the reader.

    :param reader: the command-line to parse
    :type reader: str
    :return: the reader
    :rtype: Reader
    """
    from seppl import split_args, split_cmdline, args_to_objects
    from idc.registry import available_readers

    if reader is None:
        raise Exception("No reader command-line supplied!")
    valid = dict()
    valid.update(available_readers())
    args = split_args(split_cmdline(reader), list(valid.keys()))
    objs = args_to_objects(args, valid, allow_global_options=False)
    if len(objs) == 1:
        if isinstance(objs[0], Reader):
            result = objs[0]
        else:
            raise Exception("Expected instance of Reader but got: %s" % str(type(objs[0])))
    else:
        raise Exception("Expected to obtain one reader from '%s' but got %d instead!" % (reader, len(objs)))
    return result



from typing import Tuple, List, Dict, Union
from simple_palette_utils import colors, x11_colors, COLOR_LISTS, COLOR_LIST_X11


def rgb2yiq(r: int, g: int, b: int) -> float:
    """
    Generates YIQ perceived brightness from RGB colors.

    :param r: the red value
    :type r: int
    :param g: the green value
    :type g: int
    :param b: the blue value
    :type b: int
    :return: the YIQ value
    :rtype: float
    """
    return ((r * 299) + (g * 587) + (b * 114)) / 1000


def text_color(color: Tuple[int, int, int], threshold: int = 128) -> Tuple[int, int, int]:
    """
    Computes the text color to use for the given RGB color.

    :param color: the RGB tuple (R/G/B)
    :type color: tuple
    :param threshold: the threshold to use
    :type threshold: int
    :return: the text color tuple (R/G/B)
    :rtype: tuple
    """
    r, g, b = color
    if rgb2yiq(r, g, b) >= threshold:
        return 0, 0, 0
    else:
        return 255, 255, 255



import argparse
import os
import sys
from typing import List

from seppl import get_metadata, AnyData
from seppl.io import Filter
from wai.logging import LOGGING_WARNING
from idc.api import make_list, flatten_list

MODE_INTERACTIVE = "interactive"
MODE_NONINTERACTIVE = "non-interactive"
MODES = [
    MODE_INTERACTIVE,
    MODE_NONINTERACTIVE,
]

OUTPUT_STDOUT = "stdout"
OUTPUT_STDERR = "stderr"
OUTPUT_LOGGER = "logger"
OUTPUT_FILE = "file"
OUTPUTS = [
    OUTPUT_STDOUT,
    OUTPUT_STDERR,
    OUTPUT_LOGGER,
    OUTPUT_FILE,
]


class Inspect(Filter):
    """
    Allows inspecting the data flowing through the pipeline.
    """

    def __init__(self, mode: str = MODE_INTERACTIVE, output: str = OUTPUT_STDOUT,
                 output_file: str = None, meta_data_keys: List[str] = None, show_image: bool = False,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param mode: the mode to operate in
        :type mode: str
        :param output: how to output the data
        :type output: str
        :param output_file: the file to store the data in (in case of OUTPUT_FILE)
        :type output_file: str
        :param meta_data_keys: the keys of the meta-data value to output
        :type meta_data_keys: list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)

        if mode not in MODES:
            raise Exception("Unsupported mode: %s" % mode)
        if output not in OUTPUTS:
            raise Exception("Unsupported output: %s" % output)

        self.mode = mode
        self.output = output
        self.output_file = output_file
        self.meta_data_keys = meta_data_keys
        self.show_image = show_image

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "inspect"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Allows inspecting the data flowing through the pipeline."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-m", "--mode", choices=MODES, default=MODE_INTERACTIVE, help="The mode to operate in.")
        parser.add_argument("-o", "--output", choices=OUTPUTS, default=OUTPUT_STDOUT, help="How to output the data.")
        parser.add_argument("--output_file", type=str, default=None, help="The file to store the data in, in case of output '" + OUTPUT_FILE + "'.")
        parser.add_argument("-k", "--meta-data-key", metavar="KEY", type=str, help="The meta-data value to output", required=False, nargs="*")
        parser.add_argument("-i", "--show_image", action="store_true", help="Whether to display the image.")
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.mode = ns.mode
        self.output = ns.output
        self.output_file = ns.output_file
        self.meta_data_keys = ns.meta_data_key
        self.show_image = ns.show_image

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if (self.output == OUTPUT_FILE) and ((self.output_file is None) or (self.output_file == "")):
            raise Exception("No output file provided!")
        if (self.output == OUTPUT_FILE) and (os.path.exists(self.output_file)):
            if os.path.isdir(self.output_file):
                raise Exception("Output file points to directory: %s" % self.output_file)
            os.remove(self.output_file)

    def _assemble_data(self, data) -> str:
        """
        Assembles the requested data from the record.

        :param data: the record to get the data from
        :return: the generated string
        :rtype: str
        """
        result = []

        if self.meta_data_keys is not None:
            meta = get_metadata(data)
            if meta is not None:
                for key in self.meta_data_keys:
                    if key in meta:
                        result.append("meta[%s]: %s" % (key, meta[key]))

        return "\n".join(result)

    def _output_data(self, data: str):
        """
        Outputs the requested data.
        """
        if self.output == OUTPUT_STDOUT:
            print(data)
        elif self.output == OUTPUT_STDERR:
            print(data, file=sys.stderr)
        elif self.output == OUTPUT_FILE:
            with open(self.output_file, "a") as fp:
                fp.write(data)
                fp.write("\n")
        else:
            raise Exception("Unsupported output: %s" % self.output)

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            self._output_data(self._assemble_data(item))

            if self.show_image:
                img = item.image
                if img is None:
                    self.logger().error("Failed to obtain image for display!")
                else:
                    img.show()

            if self.mode == MODE_INTERACTIVE:
                while True:
                    print("Continue (yes/no/skip)? ")
                    answer = input()
                    answer = answer.lower()
                    if (answer == "no") or (answer == "n"):
                        sys.exit(0)
                    elif (answer == "skip") or (answer == "s"):
                        break
                    elif (answer == "yes") or (answer == "y"):
                        result.append(item)
                        break
                    else:
                        print("Invalid choice!")
            else:
                result.append(item)

        return flatten_list(result)



from ._imgcls import ImageClassificationData
from ._imgseg import ImageSegmentationData
from ._objdet import ObjectDetectionData


DATATYPE_IMGCLS = "ic"
DATATYPE_IMGSEG = "is"
DATATYPE_OBJDET = "od"
DATATYPES = [
    DATATYPE_IMGCLS,
    DATATYPE_IMGSEG,
    DATATYPE_OBJDET,
]


def data_type_to_class(data_type: str):
    if data_type == DATATYPE_IMGCLS:
        return ImageClassificationData
    elif data_type == DATATYPE_IMGSEG:
        return ImageSegmentationData
    elif data_type == DATATYPE_OBJDET:
        return ObjectDetectionData
    else:
        raise Exception("Unsupported data type: %s" % data_type)



import argparse
import io
import os.path
from typing import List

from wai.logging import LOGGING_WARNING
from seppl import AnyData
from seppl.io import Filter
from idc.api import flatten_list, make_list, FORMAT_EXTENSIONS, FORMATS


class ConvertImageFormat(Filter):
    """
    Converts the image format to the specified type.
    """

    def __init__(self, image_format: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param image_format: the image format to use
        :type image_format: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.image_format = image_format

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "convert-image-format"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Converts the image format to the specified type."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-f", "--image_format", choices=FORMATS, help="The image format to use", required=True)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.image_format = ns.image_format

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.image_format is None:
            raise Exception("No image format provided!")
        if self.image_format not in FORMATS:
            raise Exception("Unsupported image format: %s" % self.image_format)

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            if item.image_format != self.image_format:
                self.logger().info("Converting %s to %s: %s" % (item.image_format, self.image_format, item.image_name))
                # convert format
                img = item.image
                output_io = io.BytesIO()
                if img.mode in ['RGBA', 'ARGB']:
                    img = img.convert('RGB')
                img.save(output_io, format=self.image_format)
                # new container
                item = type(item)(image_name=os.path.splitext(item.image_name)[0] + FORMAT_EXTENSIONS[self.image_format],
                                  data=output_io.getvalue(),
                                  metadata=item.get_metadata(), annotation=item.annotation)

            result.append(item)

        return flatten_list(result)



from ._colors import rgb2yiq, text_color
from ._fonts import DEFAULT_FONT_FAMILY, load_font, text_size
from ._data import ImageData, make_list, flatten_list, jpeg_quality, array_to_image, empty_image, save_image, FORMATS, FORMAT_JPEG, FORMAT_PNG, FORMAT_BMP, FORMAT_EXTENSIONS
from ._generator import Generator, SingleVariableGenerator
from ._imgcls import ImageClassificationData
from ._imgseg import ImageSegmentationData, ImageSegmentationAnnotations, combine_layers, split_layers
from ._imgseg import from_indexedpng, from_bluechannel, from_grayscale, to_indexedpng, to_grayscale, to_bluechannel
from ._objdet import ObjectDetectionData, get_object_label, set_object_label, DEFAULT_LABEL, LABEL_KEY
from ._utils import locate_file, locate_image, load_image_from_bytes, load_image_from_file, load_function, JPEG_EXTENSIONS
from ._utils import load_labels, save_labels, save_labels_csv
from ._utils import crop_image, pad_image
from ._utils import safe_deepcopy
from ._data_types import DATATYPE_IMGCLS, DATATYPE_OBJDET, DATATYPE_IMGSEG, DATATYPES, data_type_to_class
from ._geometry import locatedobjects_to_shapely, shapely_to_locatedobject, locatedobject_polygon_to_shapely, locatedobject_bbox_to_shapely
from ._geometry import intersect_over_union, COMBINATIONS, INTERSECT, UNION
from ._geometry import merge_polygons, fit_located_object, fit_layers
from ._reader import Reader, parse_reader
from ._filter import parse_filter
from ._writer import parse_writer
from ._writer import BatchWriter, SplittableBatchWriter
from ._writer import StreamWriter, SplittableStreamWriter
from ._writer import AnnotationsOnlyWriter, add_annotations_only_param



import argparse
import logging
import os
import traceback

from seppl import split_cmdline
from seppl.placeholders import load_user_defined_placeholders
from wai.logging import init_logging, set_logging_level, add_logging_level

from idc.api import Generator
from idc.core import ENV_IDC_LOGLEVEL
from idc.registry import available_generators
from idc.tool.convert import main as convert_main, CONVERT

EXEC = "idc-exec"

_logger = logging.getLogger(EXEC)


def execute_pipeline(pipeline: str, generator: str, dry_run: bool = False, prefix: str = False):
    """
    Executes the specified pipeline as many times as the generators produce variables.

    :param pipeline: the pipeline template to use
    :type pipeline: str
    :param generator: the generator command-line to use for generating variable values to be expanded in the pipeline template
    :type generator: str
    :param dry_run: whether to only expand/output but not execute the pipeline
    :type dry_run: bool
    :param prefix: the prefix to use when in dry-run mode
    :type prefix: str
    """
    # remove whitespaces, idc-convert from pipeline
    pipeline = pipeline.strip()
    if pipeline.startswith(CONVERT):
        pipeline = pipeline[len(CONVERT):].strip()

    # parse generator
    generator_obj = Generator.parse_generator(generator)

    # apply generator to pipeline template and execute it
    vars_list = generator_obj.generate()
    for vars_ in vars_list:
        pipeline_exp = pipeline
        for var in vars_:
            pipeline_exp = pipeline_exp.replace("{%s}" % var, vars_[var])
        _logger.info("%s\n--> %s" % (pipeline, pipeline_exp))
        if dry_run:
            if prefix is not None:
                if not prefix.endswith(" "):
                    prefix = prefix + " "
                pipeline_exp = prefix + pipeline_exp
            print(pipeline_exp)
        else:
            pipeline_args = split_cmdline(pipeline_exp)
            convert_main(pipeline_args)


def main(args=None):
    """
    The main method for parsing command-line arguments.

    :param args: the commandline arguments, uses sys.argv if not supplied
    :type args: list
    """
    init_logging(env_var=ENV_IDC_LOGLEVEL)
    generators = sorted(list(available_generators().keys()))
    parser = argparse.ArgumentParser(
        description="Tool for executing a pipeline multiple times, each time with a different set of variables expanded. A variable is surrounded by curly quotes (e.g., variable 'i' gets referenced with '{i}'). Available generators: " + ", ".join(generators),
        prog=EXEC,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-p", "--pipeline", help="The pipeline template with variables to expand and then execute.", default=None, type=str, required=True)
    parser.add_argument("-g", "--generator", help="The generator plugin to use.", default=None, type=str, required=True)
    parser.add_argument("-n", "--dry_run", action="store_true", help="Applies the generator to the pipeline template and only outputs it on stdout.", required=False)
    parser.add_argument("-P", "--prefix", help="The string to prefix the pipeline with when in dry-run mode.", required=False, default=None, type=str)
    parser.add_argument("--placeholders", metavar="FILE", help="The file with custom placeholders to load (format: key=value).", required=False, default=None, type=str)
    add_logging_level(parser)
    parsed = parser.parse_args(args=args)
    set_logging_level(_logger, parsed.logging_level)
    if parsed.placeholders is not None:
        if not os.path.exists(parsed.placeholders):
            _logger.error("Placeholder file not found: %s" % parsed.placeholders)
        else:
            _logger.info("Loading custom placeholders from: %s" % parsed.placeholders)
            load_user_defined_placeholders(parsed.placeholders)
    execute_pipeline(parsed.pipeline, parsed.generator,
                     dry_run=parsed.dry_run, prefix=parsed.prefix)


def sys_main() -> int:
    """
    Runs the main function using the system cli arguments, and
    returns a system error code.

    :return: 0 for success, 1 for failure.
    """
    try:
        main()
        return 0
    except Exception:
        traceback.print_exc()
        return 1


if __name__ == '__main__':
    main()



import argparse
import os
from typing import List

from wai.logging import LOGGING_WARNING
from idc.api import ImageClassificationData, SplittableStreamWriter, make_list
from seppl.placeholders import placeholder_list, InputBasedPlaceholderSupporter


class SubDirWriter(SplittableStreamWriter, InputBasedPlaceholderSupporter):

    def __init__(self, output_dir: str = None,
                 split_names: List[str] = None, split_ratios: List[int] = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param output_dir: the output directory to create the sub-dirs in
        :type output_dir: str
        :param split_names: the names of the splits, no splitting if None
        :type split_names: list
        :param split_ratios: the integer ratios of the splits (must sum up to 100)
        :type split_ratios: list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(split_names=split_names, split_ratios=split_ratios, logger_name=logger_name, logging_level=logging_level)
        self.output_dir = output_dir

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "to-subdir-ic"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Saves images to sub-directories, using the classification label for the sub-directory."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-o", "--output", type=str, help="The directory to create the sub-directories in according to the image labels. Any defined splits get added beneath there. " + placeholder_list(obj=self), required=True)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.output_dir = ns.output

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ImageClassificationData]

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        for item in make_list(data):
            sub_dir = self.session.expand_placeholders(self.output_dir)
            if self.splitter is not None:
                split = self.splitter.next(item=item.image_name)
                sub_dir = os.path.join(sub_dir, split)
            if item.has_annotation():
                sub_dir = os.path.join(sub_dir, item.annotation)
            if not os.path.exists(sub_dir):
                self.logger().info("Creating dir: %s" % sub_dir)
                os.makedirs(sub_dir)
            path = os.path.join(sub_dir, item.image_name)
            self.logger().info("Writing image to: %s" % path)
            item.save_image(path)



import argparse
import os
from typing import List, Iterable, Union

from wai.logging import LOGGING_WARNING
from seppl.placeholders import PlaceholderSupporter, placeholder_list
from idc.api import ImageClassificationData
from idc.api import Reader


class SubDirReader(Reader, PlaceholderSupporter):

    def __init__(self, source: Union[str, List[str]] = None, source_list: Union[str, List[str]] = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param source: the top-level directories to use
        :param source_list: the file(s) with top-level dir(s)
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.source = source
        self.source_list = source_list
        self._sub_dirs = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "from-subdir-ic"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Loads images from sub-directories, uses the name of the sub-directory as classification label."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--input", type=str, help="Path to the directory with the sub-directories containing the images; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("-I", "--input_list", type=str, help="Path to the text file(s) listing the directories to use; " + placeholder_list(obj=self), required=False, nargs="*")
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.source = ns.input
        self.source_list = ns.input_list

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ImageClassificationData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()

        # assemble top-level dirs
        all_dirs = []
        if self.source is not None:
            if isinstance(self.source, list):
                all_dirs.extend(self.source)
            else:
                all_dirs.append(self.source)
        if self.source_list is not None:
            if isinstance(self.source_list, list):
                files = self.source_list
            else:
                files = [self.source_list]
            for file in files:
                with open(file, "r") as fp:
                    lines = fp.readline()
                    lines = [x.strip() for x in lines]
                    for line in lines:
                        if len(line) == 0:
                            continue
                        if line.startswith("#"):
                            continue
                        all_dirs.append(line)

        # find all subdirs
        self._sub_dirs = dict()
        for input_dir in all_dirs:
            input_dir = self.session.expand_placeholders(input_dir)
            if not os.path.exists(input_dir):
                self.logger().warning("Directory does not exist: %s" % input_dir)
                continue
            for f in os.listdir(input_dir):
                path = os.path.join(input_dir, f)
                if os.path.isdir(path):
                    if input_dir not in self._sub_dirs:
                        self._sub_dirs[input_dir] = []
                    self._sub_dirs[input_dir].append(path)
            if input_dir not in self._sub_dirs:
                self.logger().warning("No sub-directories found in: %s" % input_dir)
            else:
                self.logger().info("Found %d sub-directories in: %s" % (len(self._sub_dirs[input_dir]), input_dir))

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        input_dirs = sorted(list(self._sub_dirs.keys()))
        for input_dir in input_dirs:
            for sub_dir in self._sub_dirs[input_dir]:
                files = []
                for f in os.listdir(sub_dir):
                    path = os.path.join(sub_dir, f)
                    f = f.lower()
                    if f.endswith(".jpg") or f.endswith(".jpeg") or f.endswith(".png"):
                        self.logger().info("Reading image from: %s" % path)
                        files.append(path)
                files = sorted(files)
                for file in files:
                    yield ImageClassificationData(source=file, annotation=os.path.basename(sub_dir))
            del self._sub_dirs[input_dir]

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return len(self._sub_dirs) == 0



from ._check_duplicate_filenames import CheckDuplicateFilenames, DUPLICATE_ACTIONS, DUPLICATE_ACTION_IGNORE, DUPLICATE_ACTION_WARN, DUPLICATE_ACTION_DROP, DUPLICATE_ACTION_ERROR
from ._convert_image_format import ConvertImageFormat
from ._discard_invalid_images import DiscardInvalidImages
from ._discard_negatives import DiscardNegatives
from ._discard_by_name import DiscardByName
from ._filter_labels import FilterLabels
from ._inspect import Inspect, MODES, MODE_NONINTERACTIVE, MODE_INTERACTIVE, OUTPUTS, OUTPUT_STDOUT, OUTPUT_STDERR, OUTPUT_LOGGER, OUTPUT_FILE
from ._map_labels import MapLabels
from ._max_records import MaxRecords
from ._metadata import MetaData
from ._metadata_from_name import MetaDataFromName
from ._metadata_objdet import MetaDataObjectDetection
from ._od_to_is import ObjectDetectionToImageSegmentation
from ._od_to_ic import ObjectDetectionToImageClassification, MULTIPLICITY, MULTIPLICITY_ERROR, MULTIPLICITY_SKIP, MULTIPLICITY_SINGLE, MULTIPLICITY_MAJORITY
from ._passthrough import PassThrough
from ._pyfunc_filter import PythonFunctionFilter
from ._randomize_records import RandomizeRecords
from ._record_window import RecordWindow
from ._remove_classes import RemoveClasses
from ._rename import Rename, RENAME_PLACEHOLDERS, RENAME_PH_NAME, RENAME_PH_COUNT, RENAME_PH_PDIR, RENAME_PH_SAME, RENAME_PH_PDIR_SUFFIX, RENAME_PH_EXT, RENAME_PH_OCCURRENCES, RENAME_PH_HELP
from ._rgb_to_grayscale import RGBToGrayscale
from ._sample import Sample
from ._set_placeholder import SetPlaceholder
from ._sort_pixels import SortPixels
from ._split_records import SplitRecords
from ._strip_annotations import StripAnnotations
from ._tee import Tee
from ._use_mask import UseMask
from ._write_labels import WriteLabels, OUTPUT_FORMATS, OUTPUT_FORMAT_TEXT, OUTPUT_FORMAT_COMMASEP, OUTPUT_FORMAT_CUSTOMSEP
